/*
 *
 *			   IPSEC for Linux
 *			 Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_ah.c,v 1.6 1998/07/14 15:52:40 rgb Exp $
 *
 * $Log: ipsec_ah.c,v $
 * Revision 1.6  1998/07/14 15:52:40  rgb
 * Point skb->dev to ipsec virtual I/F.
 *
 * Revision 1.5  1998/06/30 18:07:46  rgb
 * Change for ah/esp_protocol stuct visible only if module.
 *
 * Revision 1.4  1998/06/30 00:12:46  rgb
 * Clean up a module compile error.
 *
 * Revision 1.3  1998/06/25 19:28:06  rgb
 * Readjust premature unloading of module on packet receipt.
 * Make protocol structure abailable to rest of kernel.
 * Use macro for protocol number.
 *
 * Revision 1.2  1998/06/23 02:49:34  rgb
 * Fix minor #include bug that prevented compiling without debugging.
 * Added code to check for presence of IPIP protocol if an incoming packet
 * is IPIP encapped.
 *
 * Revision 1.1  1998/06/18 21:27:43  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid
 * kernel-build scripts happier in the presence of symlinks
 *
 * Revision 1.9  1998/06/14 23:48:43  rgb
 * Fix I/F name comparison oops bug.
 *
 * Revision 1.8  1998/06/11 07:20:02  rgb
 * Stats fixed for rx_packets.
 *
 * Revision 1.7  1998/06/11 05:53:34  rgb
 * Added stats for rx error and good packet reporting.
 *
 * Revision 1.6  1998/06/05 02:27:29  rgb
 * Add rx_errors stats.
 * Fix DoS bug:  skb's not being freed on dropped packets.
 *
 * Revision 1.5  1998/05/27 21:21:28  rgb
 * Fix DoS potential bug.  skb was not being freed if the packet was bad.
 *
 * Revision 1.4  1998/05/18 22:31:37  rgb
 * Minor change in debug output and comments.
 *
 * Revision 1.3  1998/04/21 21:29:05  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.2  1998/04/12 22:03:17  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:05:55  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:02  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * No changes.
 *
 * Revision 0.3  1996/11/20 14:35:48  ji
 * Minor Cleanup.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

#define __NO_VERSION__

#include <linux/module.h>
#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"
#ifdef CONFIG_IPSEC_AH
#include "ipsec_ah.h"
#endif

#ifdef DEBUG_IPSEC
int debug_ah = 0x0;
#endif

int 
ah_rcv(struct sk_buff *skb, struct device *dev, struct options *opt, 
       __u32 daddr, unsigned short len, __u32 saddr,
       int redo, struct inet_protocol *protocol)
{
	struct iphdr *ipp;
	struct ah *ahp;
	int iphlen;
	int i;
	unsigned char *dat;
	struct tdb *tdbp;
	struct in_addr iadst;
	struct sk_buff *skb_old = NULL;

	struct enet_statistics *stats = NULL;		/* This device's statistics */
	struct device *ipsecdev = NULL, *prvdev;
	struct ipsecpriv *prv;
	char name[9];

	/* Don't unlink in the middle of a turnaround */
	MOD_INC_USE_COUNT;
	
#ifdef DEBUG_IPSEC
	if(debug_ah)
	{
		if(!(skb->dev && dev && (skb->dev == dev)))
			printk("ah_rcv: Informational -- skb->dev:%s is not equal to dev:%s\n",
			       skb->dev->name, dev->name);
	}
#endif DEBUG_IPSEC
	for(i = 0; i < IPSEC_NUM_IF; i++)
	{
		sprintf(name, "ipsec%d", i);
		if(!strcmp(name, dev->name))
		{
			prv = (struct ipsecpriv *)(dev->priv);
			stats = (struct enet_statistics *) &(prv->mystats);
			break;
		}
		if((ipsecdev = dev_get(name)) == NULL)
			printk("ah_rcv: device %s does not exist\n", name);
		prv = ipsecdev ? (struct ipsecpriv *)(ipsecdev->priv) : NULL;
		prvdev = prv ? (struct device *)(prv->dev) : NULL;
#ifdef DEBUG_IPSEC
		if(debug_ah)
		{
			if(prvdev)
				printk("ah_rcv: physical device for device %s is %s\n",
				       name, prvdev->name);
			if(!dev)
				printk("ah_rcv: device supplied with skb is NULL\n");
		}
#endif DEBUG_IPSEC
		if(prvdev && dev && !strcmp(prvdev->name, dev->name))
		{
			stats = (struct enet_statistics *) &(prv->mystats);
			skb->dev = ipsecdev;
			stats->rx_packets++;
			break;
		}
	}
	if(!stats)
	{
		printk("ah_rcv: packet received from physical I/F not connected to ipsec I/F.  Cannot record statistics.  May not have SA for authentication.");
	}

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_REVEC)
	{
		struct timeval tv;
		do_gettimeofday(&tv);
		printk("ipsec_ah_rcv: ts=%02d.%04d AH packet received, dev = %s\n", 
		       tv.tv_sec % 60, tv.tv_usec / 100,
		       ipsecdev ? ipsecdev->name : "NULL");
	}
#endif DEBUG_IPSEC
	
	len = skb->len; dat = skb->data;
	ipp = (struct iphdr *)skb->data;
	
#ifdef INET_GET_PROTOCOL
	if((ipp->protocol == IPPROTO_IPIP) && (inet_get_protocol(IPPROTO_IPIP) == NULL))
	{
		printk("esp_rcv: error -- IPPROTO_IPIP not loaded, ICMP message will be sent back to sender.\n"); /* XXX info leak here? (rgb) */
	}
#endif

#ifdef DEBUG_IPSEC
	if (debug_ah & DB_AH_PKTRX)
	{
		printk("ipsec_ah: old ip packet is %d bytes:", len);
		ipsec_print_ip(ipp);
#ifdef DEBUG_IPSEC_AH_
		for (i = 0; i < len; i++)
			printk(" %02x", dat[i]);
		printk("\n");
#endif
	}
#endif

	iphlen = ipp->ihl << 2;
	ipp->check = 0;			/* we know the sum is good */
	ahp = (struct ah *)(skb->data + iphlen);

	/*
	 * Find tunnel control block and (indirectly) call the appropriate
	 * tranform routine. The resulting sk_buf is a valid
	 * IP packet ready to go through input processing.
	 */

	iadst.s_addr = ipp->daddr;
	tdbp = gettdb(ahp->ah_spi, iadst);
	if (tdbp == NULL)
	{
#ifdef DEBUG_IPSEC
		if (debug_ah & DB_AH_TDB)
			printk("ah_input: no tdb for spi=%x, daddr=%d.%d.%d.%d: incoming packet with no SA dropped\n",
			       (u_int)ntohl(ahp->ah_spi),
			       O4(ipp->daddr),
			       O3(ipp->daddr),
			       O2(ipp->daddr),
			       O1(ipp->daddr));
#endif
		goto rcvleave;
	}

	if (tdbp->tdb_xform == NULL)
	{
#ifdef DEBUG_IPSEC
		if (debug_ah & DB_AH_XF)
			printk("ah_input: no xform for incoming packet with spi=%x daddr=%d.%d.%d.%d, dropped\n",
			       (u_int)ntohl(ahp->ah_spi),
			       O4(ipp->daddr),
			       O3(ipp->daddr),
			       O2(ipp->daddr),
			       O1(ipp->daddr));
#endif 
		goto rcvleave;
	}

	skb_old = skb;
	skb = (*(tdbp->tdb_xform->xf_input))(skb, tdbp);
	
	if (skb == NULL)
	{
		goto rcvleave;
	}

	skb->protocol = htons(ETH_P_IP);
	skb->ip_summed = 0;
#if 0
	if(stats) /* packets encapped by esp in addition to ah will be counted twice! */
	{
		stats->rx_packets++;
	}
#endif
	netif_rx(skb);

	MOD_DEC_USE_COUNT;
	return(0);

	printk("ah_rcv: This statement should never be reached.\n");
     rcvleave:
	if(stats)
	{
		stats->rx_packets--;
		stats->rx_errors++;
	}

 	if( skb )
	{
                kfree_skb(skb, FREE_WRITE);
	}
		
 	if( skb_old )
	{
                kfree_skb(skb_old, FREE_WRITE);
	}
		
	MOD_DEC_USE_COUNT;
	return(0);
}

#ifdef CONFIG_IPSEC_MODULE
struct inet_protocol ah_protocol = {
	ah_rcv,				/* AH handler */
#if 0
	NULL,				/* Will be UDP fraglist handler */
#endif
	NULL,				/* TUNNEL error control */
	0,				/* next */
	IPPROTO_AH,			/* protocol ID */
	0,				/* copy */
	NULL,				/* data */
	"AH"				/* name */
};
#endif /* CONFIG_IPSEC_MODULE */
