/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

char ahhmacsha1_version[] = "$Id: ipsec_ahhmacsha1.c,v 1.3 1998/06/25 19:38:46 rgb Exp $";

/*
 * $Log: ipsec_ahhmacsha1.c,v $
 * Revision 1.3  1998/06/25 19:38:46  rgb
 * Check for valid authenticator size.
 * Clean up compiler warnings about argsize in printk.
 * Use macro for protocol number.
 *
 * Revision 1.2  1998/06/19 20:44:45  rgb
 * More precision demanded of user specified keys.
 * Sanity checks placed before memory allocation to save work in the case of an
 * error.
 *
 * Revision 1.8  1998/06/18 21:29:02  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid kernel
 * build scripts happier in presence of symbolic links
 *
 * Revision 1.7  1998/06/11 05:49:54  rgb
 * Added cvs version reporting.
 *
 * Revision 1.6  1998/05/27 23:19:16  rgb
 * Added version reporting.
 *
 * Revision 1.5  1998/05/21 13:14:09  rgb
 * Removed memory leak comments and other minor cleanup.
 *
 * Revision 1.4  1998/05/18 23:59:50  rgb
 * Clean up for numerical consistency of output.
 *
 * Cleaned up old debug code.  Further silenced for debug off.
 *
 * Key data is properly cleaned up, ie. zeroed.
 *
 * Comment about use of macros.
 *
 * Revision 1.3  1998/04/21 21:29:04  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.2  1998/04/12 22:03:19  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:05:57  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:06  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * New transform.
 *
 *
 */

#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#ifdef CONFIG_IPSEC_AH
#include "ipsec_ah.h"
#endif

static u_long sha1zeroes[64];

/*
 * Check-replay-window routine, adapted from the original 
 * by J. Hughes, from draft-ietf-ipsec-esp-des-md5-03.txt
 *
 *  This is a routine that implements a 64 packet window. This is intend-
 *  ed on being an implementation sample.
 */

static int
ah_checkreplaywindow(struct ahhmacsha1_xdata *xd, unsigned long seq)
{
	__u32 diff;
	
	if (seq == 0) 
		return 0;		/* first == 0 or wrapped */

	if (seq > xd->amx_lastseq)	/* new larger sequence number */
	{
		diff = seq - xd->amx_lastseq;
		if (diff < xd->amx_ooowin) /* In win, set bit for this pkt */
			xd->amx_bitmap = (xd->amx_bitmap << diff) | 1;
		else
			xd->amx_bitmap = 1; /* This packet has way larger */

		xd->amx_lastseq = seq;
		return 1;		/* larger is good */
	}
	diff = xd->amx_lastseq - seq;
	if (diff >= xd->amx_ooowin)	/* too old or wrapped */
		return 0;
	
	if (xd->amx_bitmap & (1l << diff)) /* this packet already seen */
		return 0;
	xd->amx_bitmap |= (1l << diff);	/* mark as seen */
	return 1;			/* out of order but good */
}

#ifdef DEBUG_IPSEC
static void dmp(char *s, caddr_t bb, int len)
{
	int i;
	unsigned char *b = bb;
  
	if (debug_ah & DB_AH_DMP)
	{
		printk("at %s, len=%d:", s, len);
		for (i=0; i < len; i++)
			printk(" %02x", *b++);
		printk("\n");
	}
}
#else
#define dmp(_x, _y, _z)
#endif

struct sk_buff *
ahhmacsha1_input(struct sk_buff *skb, struct tdb *tdbp)
{
	struct iphdr *ipp, ipo;
	struct ahhmacsha1_xdata *xd;
	struct ah *ahp;
	int iphlen, authlen, proto, len;
	unsigned char shaoutput[AHSHA196_ALEN];
	
#ifdef DEBUG_IPSEC
	int i;
#endif
	unsigned char *dat;
	SHA1_CTX ictx, octx; 
	
	len = skb->len;
	dat = skb->data;

	ipp = (struct iphdr *)dat;
	iphlen = ipp->ihl << 2;
	ahp = (struct ah *)(dat + iphlen);
	
	authlen = (ahp->ah_hl << 2) + ((caddr_t)&(ahp->ah_rpl) - (caddr_t)ahp);
	proto = ahp->ah_nh;
	
	xd = (struct ahhmacsha1_xdata *)tdbp->tdb_xdata;

	if (authlen != sizeof(struct ah))
	{
#ifdef DEBUG_IPSEC
		if (debug_ah & DB_AH_INAU)
			printk("ahhmacsha1_input: bad authenticator length %d, expected %d from %u.%u.%u.%u\n",
			       authlen - ((caddr_t)(ahp->ah_data) - (caddr_t)ahp), AHHMAC_HASHLEN,
			       O4((ipp->saddr)),
			       O3((ipp->saddr)),
			       O2((ipp->saddr)),
			       O1((ipp->saddr)));
#endif
	}

	ictx = xd->amx_ictx;
	octx = xd->amx_octx;;

	ipo = *ipp;
	ipo.tos = 0;
	ipo.frag_off = 0;
	ipo.ttl = 0;
	ipo.check = 0;

	dmp("ictx", (caddr_t)&ictx, sizeof(ictx));
	dmp("octx", (caddr_t)&octx, sizeof(octx));
	
	dmp("pseudoheader", (caddr_t)&ipo, sizeof (struct iphdr));
	SHA1Update(&ictx, (unsigned char *)&ipo, sizeof (struct iphdr));
	dmp("authenticator", (caddr_t)ahp, authlen - AHHMAC_HASHLEN);
	SHA1Update(&ictx, (unsigned char *)ahp, authlen - AHHMAC_HASHLEN);
	dmp("zeroes", (caddr_t)sha1zeroes, AHHMAC_HASHLEN);
	SHA1Update(&ictx, (unsigned char *)sha1zeroes, AHHMAC_HASHLEN);
	dmp("packet", (caddr_t)(dat + iphlen + authlen), len - iphlen - authlen);
	SHA1Update(&ictx,  dat + iphlen + authlen, len - iphlen - authlen);
	SHA1Final(shaoutput, &ictx);

	dmp("outer", shaoutput, AHSHA196_ALEN);
	SHA1Update(&octx, shaoutput, AHSHA196_ALEN);
	SHA1Final(shaoutput, &octx);

	if (memcmp(shaoutput, ahp->ah_data, AHHMAC_HASHLEN))
	{
#ifdef DEBUG_IPSEC
		if (debug_ah & DB_AH_INAU)
		{
			printk("ahhmacsha1_input: authentication failure on packet from %u.%u.%u.%u\n",
			       O4((ipp->saddr)),
			       O3((ipp->saddr)),
			       O2((ipp->saddr)),
			       O1((ipp->saddr)));
			dmp("got", shaoutput, AHHMAC_HASHLEN);
			dmp("shg", (caddr_t)(ahp->ah_data), AHHMAC_HASHLEN);
		}
#endif
		return NULL;
	}

	memset((caddr_t)&ictx, 0, sizeof(ictx));
	memset((caddr_t)&octx, 0, sizeof(octx));
	memset((caddr_t)shaoutput, 0, sizeof(*shaoutput));

	if (xd->amx_replayp)
	{
		unsigned long seq;

		seq = ntohl(ahp->ah_rpl);
		if (!ah_checkreplaywindow(xd, seq))
		{
#ifdef DEBUG_IPSEC
			if (debug_ah & DB_AH_REPLAY)
				printk("ahhmacsha1_input: duplicate frame from %u.%u.%u.%u\n",
				       O4((ipp->saddr)),
				       O3((ipp->saddr)),
				       O2((ipp->saddr)),
				       O1((ipp->saddr)));
#endif
			return NULL;
		}
	}

	/*
	 *	Discard the original IP header
	 */
	 
	ipp->tot_len = htons(ntohs(ipp->tot_len) - authlen);
	ipp->protocol = proto;

	memmove((void *)(skb->data + authlen), (void *)skb->data, iphlen);

	skb_pull(skb, authlen);

	/*
	 *	Adjust pointers
	 */
	 
	
	len = skb->len;
	dat = skb->data;
	
#ifdef DEBUG_IPSEC
	if (debug_ah & DB_AH_PKTRX2)
	{
		printk("ipsec_ah: valid packet forwarded to %u.%u.%u.%u is %d bytes:",
		       O4((ipp->saddr)),
		       O3((ipp->saddr)),
		       O2((ipp->saddr)),
		       O1((ipp->saddr)),
		       len);
		for (i = 0; i < len; i++)
			printk(" %02x", dat[i]);
		printk("\n");
	}
#endif

	skb->h.iph=(struct iphdr *)skb->data;
	skb->ip_hdr=(struct iphdr *)skb->data;

	memset(skb->proto_priv, 0, sizeof(struct options));
	
	ipp = (struct iphdr *)dat;
	ipp->check = ip_fast_csum((unsigned char *)dat, iphlen >> 2);

	return skb;
}

int
ahhmacsha1_attach(void)
{
#ifdef DEBUG_IPSEC
	printk("ahhmacsha1_attach: %s\n", ahhmacsha1_version);
#endif DEBUG_IPSEC
	return 0;
}

/*
 * ahhmacsha1_init() is called when an SPI is being set up. It interprets the
 * encap_msghdr present in m, and sets up the transformation data.
 */

int
ahhmacsha1_init(struct tdb *tdbp, struct xformsw *xsp, struct encap_msghdr *em)
{
	struct ahhmacsha1_edata *ed;
	struct ahhmacsha1_xdata *xd;
	unsigned char kb[AHSHA196_BLKLEN];
	
	int i;
	
	if (em->em_msglen - EMT_SETSPI_FLEN > sizeof (struct ahhmacsha1_edata))
		return EINVAL;

	ed = (struct ahhmacsha1_edata *)em->em_dat;
	
	if (ed->ame_klen != AHSHA196_KLEN)
	{
#ifdef DEBUG_IPSEC
		if(debug_ah)
			printk("ahhmacsha1_init: incorrect key size: %d -- must be %d octets (bytes)\n",
			       ed->ame_klen, AHSHA196_KLEN);
#endif DEBUG_IPSEC
		return EINVAL;
	}
	
	if (ed->ame_alen != AHSHA196_ALEN)
	{
#ifdef DEBUG_IPSEC
		if(debug_ah)
			printk("ahhmacsha1_init: incorrect authentication size: %d -- must be %d octets (bytes)\n",
			       ed->ame_alen, AHSHA196_ALEN);
#endif DEBUG_IPSEC
		return EINVAL;
	}
	
	tdbp->tdb_xform = xsp;

	tdbp->tdb_xdata = (caddr_t)kmalloc(sizeof (struct ahhmacsha1_xdata), GFP_ATOMIC);

	if (tdbp->tdb_xdata == NULL)
		return ENOBUFS;
	memset(tdbp->tdb_xdata, 0, sizeof (struct ahhmacsha1_xdata));
	
	xd = (struct ahhmacsha1_xdata *)tdbp->tdb_xdata;

	xd->amx_alen = ed->ame_alen;
	xd->amx_replayp = ed->ame_replayp;
	xd->amx_ooowin = ed->ame_ooowin;
	xd->amx_lastseq = xd->amx_bitmap = 0;
	
	SHA1Init(&xd->amx_ictx);
	SHA1Init(&xd->amx_octx);
	
	for (i = 0; i < ed->ame_klen; i++)
		kb[i] = ed->ame_key[i] ^ 0x36;
	for (; i < AHSHA196_BLKLEN; i++)
		kb[i] = 0x36;
	
	dmp("key is ", kb, AHSHA196_BLKLEN);

	SHA1Update(&xd->amx_ictx, kb, AHSHA196_BLKLEN);

	for (i = 0; i < AHSHA196_BLKLEN; i++)
		kb[i] ^= (0x36 ^ 0x5c);
	SHA1Update(&xd->amx_octx, kb, AHSHA196_BLKLEN);
	
	/* zero key buffer -- paranoid */
	memset(kb, 0, sizeof(kb));
	memset((caddr_t)&(ed->ame_key), 0, ed->ame_klen);

	return 0;
}



int ahhmacsha1_zeroize(void *xdat)
{
	struct ahhmacsha1_xdata *xd = (struct ahhmacsha1_xdata *)xdat;
	
	memset((caddr_t)xd, 0, sizeof(*xd));
	return 0;
}

int ahhmacsha1_print(void *xdat, char *buf)
{
	int size, ds;
	struct ahhmacsha1_xdata *xd = (struct ahhmacsha1_xdata *)xdat;
	size = sprintf(buf, "alen = %d, win = %d", xd->amx_alen, xd->amx_ooowin);
	buf += size;

	if (xd->amx_replayp)
	{
		ds = sprintf(buf, ", seq = %u", xd->amx_lastseq);
		buf += ds;
		size += ds;
		ds = sprintf(buf, ", bit = 0x%016Lx", xd->amx_bitmap);
		buf += ds;
		size += ds;
	}
	
	return size;
}

int
ahhmacsha1_room(struct tdb *tp, int iphlen, int tlen, int *hr, int *tr)
{
	*hr += sizeof(struct ah);  /* XXX will need to be fixed for NULL auth */
	*tr += 0;
	return 0;
}

int ahhmacsha1_output(struct sk_buff *skb, struct tdb *tdbp)
{
	struct iphdr *ipp, ipo;
	struct ahhmacsha1_xdata *xd;
	struct ah *ahp;
	int iphlen, authlen, len;
	unsigned char *dat;
	unsigned long seq;
	SHA1_CTX ictx, octx; 
	unsigned char shaoutput[AHSHA196_ALEN];
	
	ipp = (struct iphdr *)(skb->data);
	
	iphlen = ipp->ihl << 2;
	if (iphlen != sizeof (struct iphdr))
	{
		printk("ahhmacsha1_output: cannot process options yet\n"); /* XXX */
		return -1;
	}

	xd = (struct ahhmacsha1_xdata *)tdbp->tdb_xdata;

	authlen = sizeof(struct ah); /* XXX change for NULL auth */

	dat = skb_push(skb, authlen);
	len = skb->len;
	
	memmove((void *)skb->data, (void *)(skb->data + authlen), iphlen);
	ahp = (struct ah *)(dat + iphlen);

	ipp = (struct iphdr *)dat;

	if (ntohs(ipp->frag_off) & IP_OFFSET)
	{
		printk("ahhmacsha1_output: attempted to authenticate fragment!\n"); /* XXX */
		return -1;
	}

	ahp->ah_nh = ipp->protocol;
	ahp->ah_hl = (AHHMAC_HASHLEN + AHHMAC_RPLLEN) >> 2;
	ahp->ah_rv = 0;
	ahp->ah_spi = tdbp->tdb_spi;

	seq = htonl(++xd->amx_lastseq);
	ahp->ah_rpl = seq;

	ictx = xd->amx_ictx;
	octx = xd->amx_octx;
	
	ipp->protocol		=	IPPROTO_AH;
	ipp->tot_len		=	htons(skb->len);

	ipo = *ipp;
	ipo.tos = 0;
	ipo.frag_off = 0;
	ipo.ttl = 0;
	ipo.check = 0;

	dmp("ictx", (caddr_t)&ictx, sizeof(ictx));
	dmp("octx", (caddr_t)&octx, sizeof(octx));
	
	dmp("pseudoheader", (caddr_t)&ipo, sizeof (struct iphdr));
	SHA1Update(&ictx, (unsigned char *)&ipo, sizeof (struct iphdr));
	dmp("authenticator", (caddr_t)ahp, authlen - sizeof(ahp->ah_data));
	SHA1Update(&ictx, (unsigned char *)ahp, authlen - sizeof(ahp->ah_data));
	dmp("zeroes", (caddr_t)sha1zeroes, AHHMAC_HASHLEN);
	SHA1Update(&ictx, (unsigned char *)sha1zeroes, AHHMAC_HASHLEN);
	dmp("packet", (caddr_t)(dat + iphlen + authlen), len - iphlen - authlen);
	SHA1Update(&ictx,  dat + iphlen + authlen, len - iphlen - authlen);
	SHA1Final(shaoutput, &ictx);

	dmp("outer", shaoutput, AHSHA196_ALEN);
	SHA1Update(&octx, shaoutput, AHSHA196_ALEN);
	SHA1Final(shaoutput, &octx);
	
	memcpy(ahp->ah_data, shaoutput, AHHMAC_HASHLEN);

	memset((caddr_t)&ictx, 0, sizeof(ictx)); /* paranoid */
	memset((caddr_t)&octx, 0, sizeof(octx));
	memset((caddr_t)shaoutput, 0, sizeof(shaoutput));

	skb->ip_hdr = skb->h.iph = (struct iphdr *) skb->data;
	
	ipp->check = 0;
	ipp->check = ip_fast_csum((unsigned char *)ipp, ipp->ihl);

	return 0;
}
