/*
 *
			   IPSEC for Linux
		         Preliminary Release
 
	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 
		 LIMITED PRELIMINARY RELEASE LICENCE
 	
  Permission to copy, use, and distribute unmodified copies of this
  software without fee is hereby granted, provided that this entire
  notice is included in all copies.

  No modified copies may be distributed.

  [[ This restriction will, of course, change when the code becomes
  more stable. While you may of course still distribute context-diffs
  (or anything equivalent), I strongly urge you to send any changes
  you have directly to me. This will help the community by providing a
  reference base for the code. Thanks, /ji ]]
 
  THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTY. IN PARTICULAR, NEITHER THE AUTHOR NOR ANYONE
  DISTRIBUTING THIS SOFTWARE MAKE ANY REPRESENTATION OR WARRANTY OF
  ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
  FITNESS FOR ANY PARTICULAR PURPOSE.
 
 *
 */

char esp3des_version[] = "$Id: ipsec_esp3des.c,v 1.1 1998/08/08 02:52:06 rgb Exp $";

/*
 * $Log: ipsec_esp3des.c,v $
 * Revision 1.1  1998/08/08 02:52:06  rgb
 * Stripped ipsec_esp3desmd596.c to form 3des ESP without authentication.
 *
 * Revision 1.4  1998/06/25 19:41:08  rgb
 * Remove references to INITIATOR/RESPONDER which are obsolete.
 *
 * Revision 1.3  1998/06/23 02:51:53  rgb
 * Cleaned up compiler errors/warnings.
 *
 * Revision 1.2  1998/06/19 21:00:20  rgb
 * More precision demanded of user specified keys.
 * Sanity checks placed before memory allocation to save work in the case of an
 * error.
 *
 * Revision 1.9  1998/06/18 21:29:03  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid kernel
 * build scripts happier in presence of symbolic links
 *
 * Revision 1.8  1998/06/11 05:51:34  rgb
 * Added cvs version reporting.
 * Fixed self-describing padding check and put it under debug control.
 *
 * Revision 1.7  1998/06/05 02:34:21  rgb
 * Clarified the fate of packets with errors.
 *
 * Revision 1.6  1998/05/27 23:41:32  rgb
 * Added version reporting.  Separated auth and encr keys.  Print flags on
 * init.
 *
 * Revision 1.5  1998/05/21 13:14:05  rgb
 * Removed memory leak comments and other minor cleanup.
 *
 * Revision 1.4  1998/05/18 23:59:54  rgb
 * Clean up for numerical consistency of output.
 *
 * Cleaned up old debug code.  Further silenced for debug off.
 *
 * Key data is properly cleaned up, ie. zeroed.
 *
 * Comment about use of macros.
 *
 * Revision 1.3  1998/04/21 21:29:01  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.2  1998/04/12 22:03:21  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:06:02  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:06  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.5  1997/06/03 04:24:48  ji
 * New file.
 *
 *
 */

#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
/* #ifdef CONFIG_IPSEC_AH */
#include "ipsec_ah.h"
/* #endif */
/* #ifdef CONFIG_IPSEC_ESP */
#include "ipsec_esp.h"
/* #endif */

extern void des_ede3_cbc_encrypt(caddr_t, caddr_t, int, caddr_t, caddr_t, caddr_t, caddr_t, int);
extern int des_set_key(caddr_t, caddr_t);

/*
 * Check-replay-window routine, adapted from the original 
 * by J. Hughes, from draft-ietf-ipsec-esp-des-md5-03.txt
 *
 *  This is a routine that implements a 64 packet window. This is intend-
 *  ed on being an implementation sample.
 */

static int
esp_checkreplaywindow(struct esp3des_xdata *xd, unsigned long seq)
{
	unsigned long diff;
	
	/* If manual keying, don't really check replay */
	if (xd->edmx_flags & EME_MANUAL)
		return 1;

	if (seq == 0) 
		return 0;		/* first == 0 or wrapped */

	if (seq > xd->edmx_lastseq)	/* new larger sequence number */
	{
		diff = seq - xd->edmx_lastseq;
		if (diff < xd->edmx_ooowin) /* In win, set bit for this pkt */
			xd->edmx_bitmap = (xd->edmx_bitmap << diff) | 1;
		else
			xd->edmx_bitmap = 1; /* This packet has way larger */
		
		xd->edmx_lastseq = seq;
		return 1;		/* larger is good */
	}
	diff = xd->edmx_lastseq - seq;
	if (diff >= xd->edmx_ooowin)	/* too old or wrapped */ /* if wrapped, kill off SA? */
		return 0;
	
	if (xd->edmx_bitmap & (1 << diff)) /* this packet already seen */
		return 0;
	xd->edmx_bitmap |= (1 << diff);	/* mark as seen */
	return 1;			/* out of order but good */
}

struct sk_buff *
esp3des_input(struct sk_buff *skb, struct tdb *tdbp)
{
	struct iphdr *ipp;
	struct esp3des_xdata *xd;
	struct esp *espp;
	u_long seq;
	int iphlen, esphlen, proto, len, pad, ilen, padlen;
#ifdef DEBUG_IPSEC
	int i;
#endif DEBUG_IPSEC

	unsigned char *dat, *idat;
	u_long iv[2];
	
	len = skb->len;
	dat = skb->data;

	ipp = (struct iphdr *)dat;
	iphlen = ipp->ihl << 2;
	espp = (struct esp *)(dat + iphlen);
	
	xd = (struct esp3des_xdata *)tdbp->tdb_xdata;

	idat = dat + iphlen;
	ilen = len - iphlen;

#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_input: packet from %d.%d.%d.%d received with spi=0x%lx seq=%ld (iv)=0x%08lx%08lx iplen=%d esplen=%d\n",
 		       O4((ipp->saddr)),
		       O3((ipp->saddr)),
		       O2((ipp->saddr)),
		       O1((ipp->saddr)),
		       ntohl(espp->esp_spi),
		       ntohl(espp->esp_rpl),
		       ntohl(*((unsigned long *)(espp->esp_iv)    )),
		       ntohl(*((unsigned long *)(espp->esp_iv) + 1)),
		       len, ilen);
#endif
	if (ilen % 8)
	{
		printk("ipsec_esp3des_input: got packet with ilen = %d from %d.%d.%d.%d -- should be on 8 octet boundary, packet dropped\n",
		       ilen,
 		       O4((ipp->saddr)),
		       O3((ipp->saddr)),
		       O2((ipp->saddr)),
		       O1((ipp->saddr)));
		return NULL;
	}
	
	seq = ntohl(espp->esp_rpl); 
	
	if (!esp_checkreplaywindow(xd, seq))
	{
#ifdef DEBUG_IPSEC
		if (debug_esp & DB_ES_REPLAY)
			printk("esp3des_input: duplicate frame from %d.%d.%d.%d, packet dropped\n",
 		       O4((ipp->saddr)),
		       O3((ipp->saddr)),
		       O2((ipp->saddr)),
		       O1((ipp->saddr)));
#endif
		return NULL;
	}
	
	iv[0] = *((unsigned long *)(espp->esp_iv)    );
	iv[1] = *((unsigned long *)(espp->esp_iv) + 1);
	esphlen = sizeof(struct esp);

	idat += esphlen;
	ilen -= esphlen;

	des_ede3_cbc_encrypt(idat, idat, ilen,
			     (caddr_t)(xd->edmx_eks1),
			     (caddr_t)(xd->edmx_eks2),
			     (caddr_t)(xd->edmx_eks3),
			     (caddr_t)iv, 0);

	proto = idat[ilen - 1];
	padlen = idat[ilen - 2];
	pad = padlen + 2;
#ifdef DEBUG_IPSEC
	if(debug_esp & DB_ES_IPAD)
	{
		printk("esp3des_input: padlen=%d, contents: ", padlen);
		
		for (i = 1; i <= padlen; i++)
		{
			printk(" %d:%d", i, idat[ilen - 2 - padlen + i - 1]);
			if(i != idat[ilen - 2 - padlen + i - 1])
			{
				printk("esp3des_input: warning, decrypted packet from %d.%d.%d.%d has bad padding...may be bad decryption -- not dropped\n",
				       O4((ipp->saddr)),
				       O3((ipp->saddr)),
				       O2((ipp->saddr)),
				       O1((ipp->saddr)));
			} 
		}
		printk("\n");
	}
#endif

#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_IPAD)
		printk("esp3des_input: packet decrypted from %d.%d.%d.%d: proto = %d, padding = %d\n",
		       O4((ipp->saddr)),
		       O3((ipp->saddr)),
		       O2((ipp->saddr)),
		       O1((ipp->saddr)),
		       proto, pad - 2);
#endif
	
	/*
	 *	Discard the original IP header
	 */
	
	ipp->tot_len = htons(ntohs(ipp->tot_len) - esphlen - pad);
	ipp->protocol = proto;
	
	memmove((void *)(skb->data + esphlen), (void *)(skb->data), iphlen);
	
	skb_pull(skb, esphlen);
	skb_trim(skb, len - esphlen - pad);
	
	/*
	 *	Adjust pointers
	 */
	
	len = skb->len;
	dat = skb->data;
	
	skb->h.iph=(struct iphdr *)skb->data;
	skb->ip_hdr=(struct iphdr *)skb->data;

	memset(skb->proto_priv, 0, sizeof(struct options));
	
	ipp = (struct iphdr *)dat;
	ipp->check = 0;
	ipp->check = ip_fast_csum((unsigned char *)dat, iphlen >> 2);

	return skb;
}

int
esp3des_attach(void)
{
#ifdef DEBUG_IPSEC
	printk("esp3des_attach: %s\n", esp3des_version);
#endif DEBUG_IPSEC
	return 0;
}

/*
 * espdes_init() is called when an SPI is being set up. It interprets the
 * encap_msghdr present in m, and sets up the transformation data.
 */

int
esp3des_init(struct tdb *tdbp, struct xformsw *xsp, struct encap_msghdr *em)
{
	struct espblkrply_edata *ed;
	struct esp3des_xdata *xd;

	int error;

#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_init: called with tdbp=%x, xsp=%x, em=%x\n",
		       (unsigned int)tdbp, (unsigned int)xsp, (unsigned int)em);
#endif DEBUG_IPSEC
	
#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_init: called for dst=%d.%d.%d.%d, spi=0x%lx\n",
		       O4((em->em_dst.s_addr)),
		       O3((em->em_dst.s_addr)),
		       O2((em->em_dst.s_addr)),
		       O1((em->em_dst.s_addr)),
		       ntohl(em->em_spi));
#endif
	tdbp->tdb_xform = xsp;
	
	if ((em->em_msglen - EMT_SETSPI_FLEN) > sizeof (struct espblkrply_edata))
		return EINVAL;
	
	ed = (struct espblkrply_edata *)em->em_dat;
	
	if (ed->eme_klen != EMT_ESP3DES_KEY_SZ)
	{
#ifdef DEBUG_IPSEC
		if(debug_esp)
			printk("esp3des_init: incorrect encryption key size: %d -- must be %d octets (bytes)\n",
			       ed->eme_klen, EMT_ESP3DES_KEY_SZ);
#endif DEBUG_IPSEC
		return EINVAL;
	}
	
	if ( /* ed->eme_ivlen != 0 && XXX temporary until good random source found */ ed->eme_ivlen != EMT_ESPDES_IV_SZ)
	{
#ifdef DEBUG_IPSEC
		if(debug_esp)
			printk("esp3des_init: invalid IV size: %d -- must be %d octets (bytes)\n",
			       ed->eme_ivlen, EMT_ESPDES_IV_SZ);
#endif DEBUG_IPSEC
		return EINVAL;
	}
	
	tdbp->tdb_xdata = (caddr_t)kmalloc(sizeof (struct esp3des_xdata), GFP_ATOMIC);
	
	if (tdbp->tdb_xdata == NULL)
		return ENOBUFS;
	memset(tdbp->tdb_xdata, 0, sizeof (struct esp3des_xdata));
	
	xd = (struct esp3des_xdata *)tdbp->tdb_xdata;
	xd->edmx_flags = ed->eme_flags;

#ifdef DEBUG_IPSEC
	if(debug_esp)
	{
		printk("esp3des_inti: flags = 0x%x <", xd->edmx_flags);
		if (xd->edmx_flags & EME_MANUAL)
		{
			printk("MANUAL,");
		}
		/* 
		 * (other flag printing goes here)
		 */
		printk(">\n");
	}
#endif DEBUG_IPSEC

	xd->edmx_ooowin = ed->eme_ooowin;
	
#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_init: 3des key 1/3 is 0x%08lx%08lx\n",
		       ntohl(*((unsigned long *)ed->eme_key)),
		       ntohl(*((unsigned long *)ed->eme_key + 1)));
#endif DEBUG_IPSEC
	error = des_set_key((caddr_t)(ed->eme_key), (caddr_t)(xd->edmx_eks1));
	if (error == -1)
		printk("ipsec_esp3des: parity error in des key 1/3\n");
	else if (error == -2)
		printk("ipsec_esp3des: illegal weak des key 1/3\n");
	if (error)
	{
		memset(tdbp->tdb_xdata, 0, sizeof (struct esp3des_xdata));
		kfree_s(tdbp->tdb_xdata, sizeof(struct esp3des_xdata));
		return (EINVAL);
	}
	
#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_init: 3des key 2/3 is 0x%08lx%08lx\n",
		       ntohl(*((unsigned long *)ed->eme_key + 2)),
		       ntohl(*((unsigned long *)ed->eme_key + 3)));
#endif DEBUG_IPSEC
	error = des_set_key((caddr_t)(ed->eme_key) + EMT_ESPDES_KEY_SZ, (caddr_t)(xd->edmx_eks2));
	if (error == -1)
		printk("ipsec_esp3des: parity error in des key 2/3\n");
	else if (error == -2)
		printk("ipsec_esp3des: illegal weak des key 2/3\n");
	if (error)
	{
		memset(tdbp->tdb_xdata, 0, sizeof (struct esp3des_xdata));
		kfree_s(tdbp->tdb_xdata, sizeof(struct esp3des_xdata));
		return (EINVAL);
	}
	
#ifdef DEBUG_IPSEC
	if(debug_esp)
		printk("esp3des_init: 3des key 3/3 is 0x%08lx%08lx\n",
		       ntohl(*((unsigned long *)ed->eme_key + 4)),
		       ntohl(*((unsigned long *)ed->eme_key + 5)));
#endif DEBUG_IPSEC
	error = des_set_key((caddr_t)(ed->eme_key) + EMT_ESPDES_KEY_SZ * 2, (caddr_t)(xd->edmx_eks3));
	if (error == -1)
		printk("ipsec_esp3des: parity error in des key 3/3\n");
	else if (error == -2)
		printk("ipsec_esp3des: illegal weak des key 3/3\n");
	if (error)
	{
		memset(tdbp->tdb_xdata, 0, sizeof (struct esp3des_xdata));
		kfree_s(tdbp->tdb_xdata, sizeof(struct esp3des_xdata));
		return (EINVAL);
	}
	
	xd->edmx_ivlen = ed->eme_ivlen;

	xd->edmx_iv[0] = ed->eme_ivl[0];
	xd->edmx_iv[1] = ed->eme_ivl[1];
	
	xd->edmx_lastseq = 1;
	
	memset((caddr_t)&(ed->eme_key), 0, ed->eme_klen);
	
	return 0;
}

int
esp3des_zeroize(void *xdat)
{
	struct esp3des_xdata *xd = (struct esp3des_xdata *)xdat;
	
	memset((caddr_t)xd, 0, sizeof(*xd));
	return 0;
}

int
esp3des_print(void *xdat, char *buf)
{
	int size;
	int i, ds;
	union 
	{
		u_char c[8];
		u_long l[2];
	}cl;
	
	struct esp3des_xdata *xd = (struct esp3des_xdata *)xdat;
	size = sprintf(buf, "iv = 0x");
	buf += size;
	
	cl.l[0] = xd->edmx_iv[0];
	cl.l[1] = xd->edmx_iv[1];
	
	for (i = 0; i < 8; i++)
	{
		ds = sprintf(buf, "%02x", cl.c[i]);
		buf += ds;
		size += ds;
	}

	ds = sprintf(buf, " seq = %d, bit = 0x%08x, win = %d", xd->edmx_lastseq, xd->edmx_bitmap, xd->edmx_ooowin);
	buf += ds;
	size += ds;
	

	ds = sprintf(buf, " flags = 0x%x <", xd->edmx_flags);
	buf += ds;
	size += ds;

        if (xd->edmx_flags & EME_MANUAL)
        {
                ds = sprintf(buf, "MANUAL ");
                buf += ds;
                size += ds;
        }

	/* 
	 * (other flag printing goes here)
	 */

#if 0
	if (xd->edmx_flags & EME_INITIATOR)
		ds = sprintf(buf, "INITIATOR ");
	else
		ds = sprintf(buf, "RESPONDER ");
	buf += ds;
	size += ds;
#endif /* 0 */
	
	ds = sprintf(buf, ">");
	buf += ds;
	size += ds;

	return size;
}

int
esp3des_room(struct tdb *tp, int iphlen, int tlen, int *hr, int *tr)
{
	
	*hr += sizeof(struct esp); 
	*tr += ((8 - ((tlen + 2) % 8)) % 8) + 2; /* padding */
	
	return 0;
}

int
esp3des_output(struct sk_buff *skb, struct tdb *tdbp)
{
	struct iphdr *ipp;
	struct esp3des_xdata *xd;
	struct esp *espp;
	int iphlen, esphlen, hrm, trm, len, ilen, padlen, i;
	unsigned char *dat, *idat, *pad;
	__u32 iv[2];
	
	ipp = (struct iphdr *)(skb->data);
#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_OINFO)
	{
		printk("esp3des_output: old header (headroom = %d, tailroom = %d\n",
		       skb_headroom(skb), skb_tailroom(skb));
	}
#endif	
	iphlen = ipp->ihl << 2;
	if (iphlen != sizeof (struct iphdr))
	{
		printk("esp3des_output: cannot process IP header options yet\n"); /* XXX */
		return -1;
	}
	
	xd = (struct esp3des_xdata *)(tdbp->tdb_xdata);
	
	/* see esp3des_room() for next two lines */
	hrm = sizeof(struct esp);
	trm = ((8 - ((skb->len - iphlen + 2) % 8)) % 8)	+ 2;
	
	dat = skb_push(skb, hrm);
	ilen = skb->len;
	len = ilen + trm;
	
#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_OINFO2)
		printk("esp3des_output: before skb_trim: %d,%d (need: %d,%d)",
		       skb_headroom(skb), skb_tailroom(skb), hrm, trm);
#endif
	skb_put(skb, trm);
#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_OINFO2)
		printk(" after skb_trim: %d,%d\n", skb_headroom(skb), skb_tailroom(skb));
#endif	
	memmove((void *)dat, (void *)(dat + hrm), iphlen);
	espp = (struct esp *)(dat + iphlen);
	ipp = (struct iphdr *)dat;

	if (ntohs(ipp->frag_off) & IP_OFFSET)
	{
		printk("esp3des_output: attempted to encrypt fragment!\n");
		return -1;
	}
	
	espp->esp_spi = tdbp->tdb_spi;
	espp->esp_rpl = htonl(xd->edmx_lastseq++);
	esphlen = sizeof(struct esp);		/* what's not encrypted */
	iv[0] = *((unsigned long *)&(espp->esp_iv)    ) = xd->edmx_iv[0];
	iv[1] = *((unsigned long *)&(espp->esp_iv) + 1) = xd->edmx_iv[1];

	/* Self-describing padding */
	pad = &dat[len - trm];
	padlen = trm - 2;
	for (i = 0; i < padlen; i++)
		pad[i] = i + 1;

	dat[len - 2] = padlen;
	dat[len - 1] = ipp->protocol;

	idat = dat + iphlen + esphlen;
	ilen = len - iphlen - esphlen;

#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_OINFO)
		printk("esp3des_output: encrypting %d bytes\n", ilen);
#endif	
	des_ede3_cbc_encrypt(idat, idat, ilen,
			     (caddr_t)(xd->edmx_eks1),
			     (caddr_t)(xd->edmx_eks2),
			     (caddr_t)(xd->edmx_eks3),
			     (caddr_t)iv, 1);

	/* XXX update IV with the last 8 octets of the encryption */
	xd->edmx_iv[0] = ((__u32 *)(idat))[(ilen >> 2) - 2];
	xd->edmx_iv[1] = ((__u32 *)(idat))[(ilen >> 2) - 1];
	
	ipp->protocol		=	IPPROTO_ESP; /* should come from /usr/src/linux/include/linux/in.h ? */
	ipp->tot_len		=	htons(len);
	
	skb->ip_hdr = skb->h.iph = (struct iphdr *) skb->data;
	ipp->check = 0;
	ipp->check = ip_fast_csum((unsigned char *)ipp, ipp->ihl);
	
#ifdef DEBUG_IPSEC
	if (debug_esp & DB_ES_OH) {
		printk("esp3des_output: new header (headroom = %d, tailroom = %d\n",
		       skb_headroom(skb), skb_tailroom(skb));
		printk("esp3des_output: ");
		ipsec_print_ip(ipp);
	}
#endif
	
	return 0;
}
