/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

char ipsec_init_c_version[] = "$Id: ipsec_init.c,v 1.4 1998/07/27 21:50:22 rgb Exp $";

/*
 * $Log: ipsec_init.c,v $
 * Revision 1.4  1998/07/27 21:50:22  rgb
 * Not necessary to traverse mask tree for /proc/net/ipsec_eroute.
 *
 * Revision 1.3  1998/06/25 19:51:20  rgb
 * Clean up #endif comments.
 * Shift debugging comment control for procfs to debug_tunnel.
 * Make proc_dir_entries visible to rest of kernel for static link.
 * Replace hardwired fileperms with macros.
 * Use macros for procfs inode numbers.
 * Rearrange initialisations between ipsec_init and module_init as appropriate
 * for static loading.
 *
 * Revision 1.2  1998/06/23 02:55:43  rgb
 * Slightly quieted init-time messages.
 * Re-introduced inet_add_protocol after it mysteriously disappeared...
 * Check for and warn of absence of IPIP protocol on install of module.
 * Move tdbcleanup to ipsec_xform.c.
 *
 * Revision 1.10  1998/06/18 21:29:04  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid kernel
 * build scripts happier in presence of symbolic links
 *
 * Revision 1.9  1998/06/14 23:49:40  rgb
 * Clarify version reporting on module loading.
 *
 * Revision 1.8  1998/06/11 05:54:23  rgb
 * Added /proc/net/ipsec_version to report freeswan and transform versions.
 * Added /proc/net/ipsec_spinew to generate new and unique spi's..
 * Fixed /proc/net/ipsec_tncfg bug.
 *
 * Revision 1.7  1998/05/25 20:23:13  rgb
 * proc_register changed to dynamic registration to avoid arbitrary inode
 * numbers.
 *
 * Implement memory recovery from tdb and eroute tables.
 *
 * Revision 1.6  1998/05/21 13:08:58  rgb
 * Rewrote procinfo subroutines to avoid *bad things* when more that 3k of
 * information is available for printout.
 *
 * Revision 1.5  1998/05/18 21:29:48  rgb
 * Cleaned up /proc/net/ipsec_* output, including a title line, algorithm
 * names instead of numbers, standard format for numerical output base,
 * whitespace for legibility, and the names themselves for consistency.
 *
 * Added /proc/net/ipsec_spigrp and /proc/net/ipsec_tncfg.
 *
 * Revision 1.4  1998/04/30 15:42:24  rgb
 * Silencing attach for normal operations with #ifdef IPSEC_DEBUG.
 *
 * Revision 1.3  1998/04/21 21:28:58  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.2  1998/04/12 22:03:22  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:06:05  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:02  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * No changes.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Fixed problem with node names of /proc/net entries.
 * Other minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

/*
 * Initialization code, and /proc file system interface code.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"

#include "version.c"

#ifdef CONFIG_IPSEC_AH
#include "ipsec_ah.h"
#endif /* CONFIG_IPSEC_AH */

#ifdef CONFIG_IPSEC_ESP
#include "ipsec_esp.h"
#endif /* CONFIG_IPSEC_ESP */

#ifdef DEBUG_IPSEC
int debug_eroute = 0;
int debug_spi = 0;
#endif /* DEBUG_IPSEC */

unsigned int ipsec_spi_next = 0x1000;

static int
ipsec_eroute_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{
	struct wsbuf w = {buffer, length, offset, 0, 0, 0, 0};

#ifdef DEBUG_IPSEC
	if (debug_radij & DB_RJ_DUMPTREES)
	  rj_dumptrees();			/* XXXXXXXXX */
#endif /* DEBUG_IPSEC */

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
	  printk("ipsec_eroute_get_info: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n", (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	w.len = sprintf(buffer, "EROUTE:\n"); 

	rj_walktree(rnh, ipsec_rj_walker_procprint, &w);
/*	rj_walktree(mask_rjhead, ipsec_rj_walker_procprint, &w); */

	*start = buffer + (offset - w.begin);	/* Start of wanted data */
	w.len -= (offset - w.begin);			/* Start slop */
	if (w.len > length)
		w.len = length;
	return w.len;
}

static int
ipsec_spi_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{

	int len = 0, prev_len = 0;
	off_t pos = 0, begin = 0;
	int i;
	struct tdb *tdbp;

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
	  printk("ipsec_spi_get_info: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n", (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	len = sprintf(buffer, "SPI:\n"); 

	for (i = 0; i < TDB_HASHMOD; i++)
		for (tdbp = tdbh[i]; tdbp; tdbp = tdbp->tdb_hnext)
		{
			len += sprintf(buffer + len, "(%d.%d.%d.%d, 0x%x, %s: ", 
					O4(tdbp->tdb_dst.s_addr),
					O3(tdbp->tdb_dst.s_addr),
					O2(tdbp->tdb_dst.s_addr),
					O1(tdbp->tdb_dst.s_addr),
					(u_int)ntohl(tdbp->tdb_spi),
					tdbp->tdb_xform->xf_name);
			len += tdbp->tdb_xform->xf_print(tdbp->tdb_xdata, buffer + len);
			len += sprintf(buffer + len, ")\n\n");
			buffer[len] = '\0';
			pos = begin + len;
			if(pos < offset)
			{
				len = 0;
				begin = pos;
			}
			else if (pos > offset + length)
			{
				len = prev_len;
				goto done_spi_i;
			}
			prev_len = len;
		}
 done_spi_i:	
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;

}

static int
ipsec_spigrp_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{

	int len = 0, prev_len = 0;
	off_t pos = 0, begin = 0;
	int i;
	struct tdb *tdbp, *tdbp2;

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
	  printk("ipsec_spigrp_get_info: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n", (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	len = sprintf(buffer, "SPIGRP:\n"); 

	for (i = 0; i < TDB_HASHMOD; i++)
		for (tdbp = tdbh[i]; tdbp; tdbp = tdbp->tdb_hnext)
		{
			if(!tdbp->tdb_inext)
			{
				tdbp2 = tdbp;
				while(tdbp2)
				{
					len += sprintf(buffer + len, "(%d.%d.%d.%d, 0x%x, %s", 
						     O4(tdbp2->tdb_dst.s_addr),
						     O3(tdbp2->tdb_dst.s_addr),
						     O2(tdbp2->tdb_dst.s_addr),
						     O1(tdbp2->tdb_dst.s_addr),
						     (u_int)ntohl(tdbp2->tdb_spi),
						     tdbp2->tdb_xform->xf_name);
#if 0
					len += sprintf(buffer + len, ": ");
					len += tdbp2->tdb_xform->xf_print(tdbp2->tdb_xdata, buffer + len);
#endif
					len += sprintf(buffer + len, ")\n");
					tdbp2 = tdbp2->tdb_onext;
				}
				len += sprintf(buffer + len, "\n");
				buffer[len] = '\0';
				pos = begin + len;
				if(pos < offset)
				{
					len = 0;
					begin = pos;
				}
				else if (pos > offset + length)
				{
					len = prev_len;
					goto done_spigrp_i;
				}
				prev_len = len;
			}
		}
 done_spigrp_i:	
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;
}

static int
ipsec_tncfg_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{
	int len = 0, prev_len = 0;
	off_t pos = 0, begin = 0;
	int i;
	char name[9];
	struct device *dev, *privdev;
	struct ipsecpriv *priv;

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
		printk("ipsec_tncfg_get_info: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n",
		       (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	len = sprintf(buffer, "TNCFG:\n");

	for(i = 0; i < IPSEC_NUM_IF; i++)
	{
		sprintf(name, "ipsec%d", i);
		dev = dev_get(name);
		if(dev)
		{
			priv = (struct ipsecpriv *)(dev->priv);
			privdev = (struct device *)(priv->dev);
			len += sprintf(buffer + len, "%s -> %s\n",
				       dev->name, privdev ? privdev->name : "NULL");
			buffer[len] = '\0';
			pos = begin + len;
			if(pos < offset)
			{
				len = 0;
				begin = pos;
			}
			else if (pos > offset + length)
			{
				len = prev_len;
				break;
			}
			prev_len = len;
		}
	}
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;
}

static int
ipsec_spi_get_new(char *buffer, char **start, off_t offset, int length, int dummy)
{

	int len = 0, prev_len = 0;
	off_t pos = 0, begin = 0;

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
	  printk("ipsec_spi_get_new: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n", (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	len = sprintf(buffer, "SPINEW:\n");
	len += sprintf(buffer + len, "0x%x\n", ipsec_spi_next++);

	buffer[len] = '\0';
	pos = begin + len;
	if(pos < offset)
	{
		len = 0;
		begin = pos;
	}
	else if (pos > offset + length)
	{
		len = prev_len;
		goto done_spinew;
	}
	prev_len = len;
	
 done_spinew:	
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;

}

static int
ipsec_version_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{

	int len = 0, prev_len = 0;
	off_t pos = 0, begin = 0;
	struct xformsw *xsp;

#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
	  printk("ipsec_version_get_info: buffer=0x%x, *start=0x%x, offset=%d, length=%d\n", (u_int)buffer, (u_int)*start, (int)offset, length);
#endif /* DEBUG_IPSEC */

	len = sprintf(buffer, "VERSION_INFO:\n");
	len += sprintf(buffer + len, "FreeS/WAN version: %s\n", freeswan_version);
#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_PROCFS)
		printk("ipsec_init CVS version %s\n", ipsec_init_c_version);
#endif /* DEBUG_IPSEC */

	for (xsp = xformsw; xsp < xformswNXFORMSW; xsp++)
	{
		if(xsp->xf_name)
			len += sprintf(buffer + len, "Transform: <%s>, ", xsp->xf_name);
		
		if(xsp->xf_cvsid)
			len += sprintf(buffer + len, "CVS version: %s", xsp->xf_cvsid);

		len += sprintf(buffer + len, "\n");

		buffer[len] = '\0';
		pos = begin + len;
		if(pos < offset)
		{
			len = 0;
			begin = pos;
		}
		else if (pos > offset + length)
		{
			len = prev_len;
			goto done_version_i;
		}
		prev_len = len;
	}
	
 done_version_i:	
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;

}

struct proc_dir_entry ipsec_eroute =
{
	PROC_NET_IPSEC_EROUTE,
	12, "ipsec_eroute",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_eroute_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_spi =
{
	PROC_NET_IPSEC_SPI,
	9, "ipsec_spi",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_spi_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_spigrp =
{
	PROC_NET_IPSEC_SPIGRP,
	12, "ipsec_spigrp",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_spigrp_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_tncfg =
{
	PROC_NET_IPSEC_TNCFG,
	11, "ipsec_tncfg",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_tncfg_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_spinew =
{
	PROC_NET_IPSEC_SPINEW,
	12, "ipsec_spinew",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_spi_get_new,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_version =
{
	PROC_NET_IPSEC_VERSION,
	13, "ipsec_version",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_version_get_info,
	NULL, NULL, NULL, NULL, NULL
};

void
ipsec_init(void)
{
	struct xformsw *xsp;

	(void)netlink_attach(NETLINK_IPSEC, ipsec_callback);

	ipsec_radijinit();

#ifdef INET_GET_PROTOCOL
	if(inet_get_protocol(IPPROTO_IPIP) == NULL)
	{
		printk("ipsec_init: warning -- IPPROTO_IPIP not loaded, tunnel mode will not receive.\n");
	}
#endif /* INET_GET_PROTOCOL */

	for (xsp = xformsw; xsp < xformswNXFORMSW; xsp++)
	{
#ifdef DEBUG_IPSEC_
		printk("ipsec_init: attaching <%s>\n", xsp->xf_name);
#endif /* DEBUG_IPSEC */
		(*(xsp->xf_attach))();
	}
}	


void ipsec_cleanup(void)
{
	netlink_detach(NETLINK_IPSEC);

	ipsec_tdbcleanup();
	ipsec_radijcleanup();
}

#ifdef MODULE
int init_module(void)
{
	proc_net_register(&ipsec_eroute);
	proc_net_register(&ipsec_spi);
	proc_net_register(&ipsec_spigrp);
	proc_net_register(&ipsec_tncfg);
	proc_net_register(&ipsec_spinew);
	proc_net_register(&ipsec_version);

	ipsec_init();

#ifdef CONFIG_IPSEC_ESP
	inet_add_protocol(&esp_protocol);
#endif /* CONFIG_IPSEC_ESP */

#ifdef CONFIG_IPSEC_AH
	inet_add_protocol(&ah_protocol);
#endif /* CONFIG_IPSEC_AH */

	printk("ipsec_init: ipsec module loaded. freeswan version: %s\n", freeswan_version);
	ipsec_tunnel_init_module();
	return 0;
}

int cleanup_module(void)
{
	ipsec_tunnel_cleanup_module();

#ifdef CONFIG_IPSEC_AH
	if ( inet_del_protocol(&ah_protocol) < 0 )
		printk(KERN_INFO "ah close: can't remove protocol\n");
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_ESP
	if ( inet_del_protocol(&esp_protocol) < 0 )
		printk(KERN_INFO "esp close: can't remove protocol\n");
#endif /* CONFIG_IPSEC_ESP */

	if (proc_net_unregister(ipsec_version.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-version\n");
	if (proc_net_unregister(ipsec_spinew.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-spinew\n");
	if (proc_net_unregister(ipsec_eroute.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-tncfg\n");
	if (proc_net_unregister(ipsec_spi.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-spigrp\n");
	if (proc_net_unregister(ipsec_spigrp.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-spi\n");
	if (proc_net_unregister(ipsec_tncfg.low_ino) != 0)
		printk("ipsec_cleanup: cannot unregister /proc/net/ipsec-eroute\n");

	ipsec_cleanup();

	printk("ipsec_cleanup: ipsec module unloaded.\n");
#ifdef DEBUG_IPSEC
	if (debug_tunnel & DB_TN_INIT)
		printk("ipsec_tunnel_cleanup_module: unloaded module ipsec\n");
#endif /* DEBUG_IPSEC */
	return 0;
}
#endif /* MODULE */
