/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_netlink.c,v 1.7 1998/08/12 00:10:06 rgb Exp $
 *
 * $Log: ipsec_netlink.c,v $
 * Revision 1.7  1998/08/12 00:10:06  rgb
 * Fixed minor error return code syntax.
 *
 * Revision 1.6  1998/07/29 20:22:57  rgb
 * Cosmetic cleanup.
 *
 * Revision 1.5  1998/07/27 21:53:11  rgb
 * Check for proper return code from eroute clear command.
 * Use appropriate error return codes from kernel.
 * Add an option to clear the SA table.
 *
 * Revision 1.4  1998/07/14 18:02:40  rgb
 * Add a command to clear the eroute table.
 * Clean up some error codes.
 *
 * Revision 1.3  1998/06/25 19:52:33  rgb
 * Code cosmetic changes only.
 *
 * Revision 1.2  1998/06/23 02:57:58  rgb
 * Clean up after an error condition in setspi.
 *
 * Revision 1.9  1998/06/18 21:29:06  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid kernel
 * build scripts happier in presence of symbolic links
 *
 * Revision 1.8  1998/06/08 17:57:15  rgb
 * Very minor spacing change.
 *
 * Revision 1.7  1998/05/18 21:46:45  rgb
 * Clean up for numerical consistency of output.
 *
 * Added debugging switch output.
 *
 * SETSPI will refuse to overwrite a previous SA.  This is to make it
 * consistent with the eroute command.
 *
 * spidel now deletes entire chain of spi's.
 *
 * spigrp can now ungroup a set of spi's.
 *
 * spigrp will not regroup a previously grouped spi.
 *
 * Key data is properly cleaned up, ie. zeroed.
 *
 * Revision 1.6  1998/05/07 20:36:27  rgb
 * Fixed case where debugging not enabled that caused ipsec_netlink.c to
 * not compile.
 *
 * Revision 1.5  1998/05/06 03:34:21  rgb
 * Updated debugging output statements.
 *
 * Revision 1.4  1998/04/23 21:03:59  rgb
 * Completed kernel development for userspace access to klips kernel debugging
 * switches.
 * Added detail to the kernel error message when trying to group non-existant
 * spi's.
 *
 * Revision 1.3  1998/04/21 21:29:06  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.2  1998/04/12 22:03:23  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:06:08  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:02  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * No changes.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>
#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"

#ifdef CONFIG_IPSEC_AH
#include "ipsec_ah.h"
#endif

#ifdef CONFIG_IPSEC_ESP
#include "ipsec_esp.h"
#endif

#ifdef DEBUG_IPSEC
#include "ipsec_tunnel.h"
#endif

#ifdef DEBUG_IPSEC
int debug_netlink = 0;
#endif

/*
 * IPSEC <> netlink interface.
 */

#define SENDERR(_x) do { len = -(_x); goto errlab; } while (0)

int 
ipsec_callback(struct sk_buff *skb)
{
	/*
	 * this happens when we write to /dev/ipsec (c 36 10)
	 */
	int len = skb->len;
	u_char *dat = (u_char *)skb->data;
	struct encap_msghdr *em	= (struct encap_msghdr *)dat;
	struct tdb *tdbp, *tprev;
	int i, nspis, error;
#ifdef DEBUG_IPSEC
	struct eroute *eret;
	if(debug_netlink)
		printk("ipsec_callback: skb=%x skblen=%ld em_magic=%d em_type=%d em_spi=0x%lx\n",
		       (unsigned int)skb, skb->len, em->em_magic, em->em_type, ntohl(em->em_spi));
#endif	
	/*	em = (struct encap_msghdr *)dat; */
	if (em->em_magic != EM_MAGIC) {
		return -EINVAL;
	}
	switch (em->em_type) {
	case EMT_SETDEBUG:
#ifdef DEBUG_IPSEC
		if(em->em_db_nl >> (sizeof(em->em_db_nl) * 8 - 1)) {
			if(debug_netlink)
				printk("ipsec_callback: set\n");
			debug_tunnel  |= em->em_db_tn;
			debug_netlink |= em->em_db_nl;
			debug_xform   |= em->em_db_xf;
			debug_eroute  |= em->em_db_er;
			debug_spi     |= em->em_db_sp;
			debug_radij   |= em->em_db_rj;
			debug_esp     |= em->em_db_es;
			debug_ah      |= em->em_db_ah;
		} else {
			if(debug_netlink)
				printk("ipsec_callback: unset\n");
			debug_tunnel  &= em->em_db_tn;
			debug_netlink &= em->em_db_nl;
			debug_xform   &= em->em_db_xf;
			debug_eroute  &= em->em_db_er;
			debug_spi     &= em->em_db_sp;
			debug_radij   &= em->em_db_rj;
			debug_esp     &= em->em_db_es;
			debug_ah      &= em->em_db_ah;
		}
		break;
#else
		printk("ipsec_callback: debugging not enabled\n");
		SENDERR(EINVAL);
#endif	

	case EMT_SETEROUTE:
		if (!ipsec_makeroute(&(em->em_eaddr), &(em->em_emask), em->em_erdst, em->em_erspi))
			SENDERR(EINVAL);
		break;

	case EMT_DELEROUTE:
		if (!ipsec_breakroute(&(em->em_eaddr), &(em->em_emask)))
			SENDERR(EINVAL);
		break;

	case EMT_CLREROUTE:
		if (ipsec_cleareroutes())
			SENDERR(EINVAL); /* XXX wrong error code */
		break;

#ifdef DEBUG_IPSEC
	case EMT_TESTROUTE:
		      
		printk("ipsec_callback: testroute: trying to locate 0x%08x->0x%08x\n", (u_int)em->em_eaddr.sen_ip_src.s_addr, (u_int)em->em_eaddr.sen_ip_dst.s_addr);

		eret = ipsec_findroute(&(em->em_eaddr));
		
		if (eret == NULL) {
			printk(" not found\n");
		} else {
			printk(" via %u.%u.%u.%u, 0x%x\n",
			       O4(eret->er_dst.s_addr),
			       O3(eret->er_dst.s_addr),
			       O2(eret->er_dst.s_addr),
			       O1(eret->er_dst.s_addr),
			       (u_int)ntohl(eret->er_spi));
		}
		break;
#endif
	case EMT_SETSPI:
		if (em->em_if >= 5)	/* XXX -- why 5? */
			SENDERR(ENODEV);
		
		tdbp = gettdb(em->em_spi, em->em_dst);
		if (tdbp == NULL) {
#ifdef DEBUG_IPSEC
			if (debug_netlink & DB_NL_TDBCB)
				printk("ipsec_callback: could not find a TDB for spi=0x%x, daddr=%u.%u.%u.%u, allocating (this is normal)\n",
				       (u_int)ntohl(em->em_spi),
				       O4((em->em_dst.s_addr)),
				       O3((em->em_dst.s_addr)),
				       O2((em->em_dst.s_addr)),
				       O1((em->em_dst.s_addr)));
#endif
			tdbp = (struct tdb *)kmalloc(sizeof (*tdbp), GFP_ATOMIC);

			if (tdbp == NULL)
				SENDERR(ENOBUFS);
			
			memset((caddr_t)tdbp, 0, sizeof(*tdbp));
			
			tdbp->tdb_spi = em->em_spi;
			tdbp->tdb_dst = em->em_dst;
/* XXX  		tdbp->tdb_rcvif = &(enc_softc[em->em_if].enc_if);*/
			tdbp->tdb_rcvif = NULL;
			puttdb(tdbp);
		} else {
#ifdef DEBUG_IPSEC
			if (debug_netlink & DB_NL_TDBCB)
				printk("ipsec_callback: EMT_SETSPI found a TDB for 0x%x, %u.%u.%u.%u: must delete old SA first!\n",
				       (u_int)ntohl(em->em_spi),
				       O4((em->em_dst.s_addr)),
				       O3((em->em_dst.s_addr)),
				       O2((em->em_dst.s_addr)),
				       O1((em->em_dst.s_addr)));
#endif
			SENDERR(EEXIST);
		}
		
		if ((error = tdb_init(tdbp, em))) {
			if (tdbp->tdb_xdata) {
				tdbp->tdb_xform->xf_zeroize(tdbp->tdb_xdata);
				kfree((caddr_t)tdbp->tdb_xdata);
				tdbp->tdb_xdata = NULL;
			}
			deltdb(tdbp);
			memset((caddr_t)tdbp, 0, sizeof(*tdbp));
			kfree_s(tdbp, sizeof(struct (*tdbp)));
			
			SENDERR(error);
		}
		break;
		
	case EMT_DELSPI:
		if (em->em_if >= 5)	/* XXX -- why 5? */
			SENDERR(ENODEV);
		
		tdbp = gettdb(em->em_spi, em->em_dst);
		if (tdbp == NULL) {
#ifdef DEBUG_IPSEC
			if (debug_netlink & DB_NL_TDBCB)
				printk("ipsec_callback: EMT_DELSPI could not find a TDB for 0x%x, %u.%u.%u.%u to delete!\n",
				       (u_int)ntohl(em->em_spi),
				       O4((em->em_dst.s_addr)),
				       O3((em->em_dst.s_addr)),
				       O2((em->em_dst.s_addr)),
				       O1((em->em_dst.s_addr)));
#endif
			SENDERR(ENXIO);  /* XXX -- wrong error message... */
		} else {
			while(tdbp->tdb_onext) {
				tdbp = tdbp->tdb_onext;
			}
			while(tdbp) {
				tprev = tdbp;
				tdbp = tdbp->tdb_inext;
				tprev->tdb_inext = NULL;
				if(tdbp)
					tdbp->tdb_onext = NULL;
				if (tprev->tdb_xdata) {
					tprev->tdb_xform->xf_zeroize(tprev->tdb_xdata);
					kfree((caddr_t)tprev->tdb_xdata);
					tprev->tdb_xdata = NULL;
				}
				deltdb(tprev);
				memset((caddr_t)tprev, 0, sizeof(*tprev));
				kfree_s(tprev, sizeof(struct (*tprev)));
			}
		}
		break;
		
	case EMT_GRPSPIS:
		nspis = (len - 8) / 12;
		if (nspis * 12 + 8 != len) {
			SENDERR(EINVAL);
			break;
		}
		
		for (i = 0; i < nspis; i++)
			if ((tdbp = gettdb(em->em_rel[i].emr_spi, em->em_rel[i].emr_dst)) == NULL) {
				printk("ipsec_callback: groupspis: no TDB for %u.%u.%u.%u, 0x%x\n",
				       O4((em->em_rel[i].emr_dst.s_addr)),
				       O3((em->em_rel[i].emr_dst.s_addr)),
				       O2((em->em_rel[i].emr_dst.s_addr)),
				       O1((em->em_rel[i].emr_dst.s_addr)),
				       (u_int)ntohl(em->em_rel[i].emr_spi)); /* XXX */
				printk("ipsec_callback: meaning -- tried to group a non-existant SPI\n");
				SENDERR(ENXIO);
			} else {
				if(tdbp->tdb_inext || tdbp->tdb_onext) {
					printk("ipsec_callback: groupspis: TDB for %u.%u.%u.%u, 0x%x already grouped\n",
					       O4((em->em_rel[i].emr_dst.s_addr)),
					       O3((em->em_rel[i].emr_dst.s_addr)),
					       O2((em->em_rel[i].emr_dst.s_addr)),
					       O1((em->em_rel[i].emr_dst.s_addr)),
					       (u_int)ntohl(em->em_rel[i].emr_spi)); /* XXX */
					printk("ipsec_callback: groupspis: meaning -- tried to group a previously grouped SPI\n");
					SENDERR(EBUSY);
				}
				em->em_rel[i].emr_tdb = tdbp;
			}
		tprev = em->em_rel[0].emr_tdb;
		tprev->tdb_inext = NULL;
		for (i = 1; i < nspis; i++) {
			tdbp = em->em_rel[i].emr_tdb;
			tprev->tdb_onext = tdbp;
			tdbp->tdb_inext = tprev;
			tprev = tdbp;
		}
		tprev->tdb_onext = NULL;
		error = 0;
		break;
		
	case EMT_UNGRPSPIS:
		if (len != 12) {
			SENDERR(EINVAL);
			break;
		}
		
		if ((tdbp = gettdb(em->em_rel[0].emr_spi, em->em_rel[0].emr_dst)) == NULL) {
			printk("ipsec_callback: ungroupspis: no TDB for %u.%u.%u.%u, 0x%x\n",
			       O4((em->em_rel[0].emr_dst.s_addr)),
			       O3((em->em_rel[0].emr_dst.s_addr)),
			       O2((em->em_rel[0].emr_dst.s_addr)),
			       O1((em->em_rel[0].emr_dst.s_addr)),
			       (u_int)ntohl(em->em_rel[0].emr_spi)); /* XXX */
			printk("ipsec_callback: meaning -- tried to ungroup a non-existant SPI\n");
			SENDERR(ENXIO);
		}
		while(tdbp->tdb_onext) {
			tdbp = tdbp->tdb_onext;
		}
		while(tdbp->tdb_inext) {
			tprev = tdbp;
			tdbp = tdbp->tdb_inext;
			tprev->tdb_inext = NULL;
			tdbp->tdb_onext = NULL;
		}

		error = 0;
		break;
		
	case EMT_CLRSPIS:
#ifdef DEBUG_IPSEC
		if(debug_netlink) {
			printk("ipsec_callback: spi clear called.\n");
		}
#endif	
		if (em->em_if >= 5)	/* XXX -- why 5? */
			SENDERR(ENODEV);
		ipsec_tdbcleanup();
		break;
	default:
		printk("ipsec_callback: unknown message type\n");
	}
 errlab:
	return len;
}
