/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_netlink.h,v 1.3 1998/07/27 21:54:22 rgb Exp $
 *
 * $Log: ipsec_netlink.h,v $
 * Revision 1.3  1998/07/27 21:54:22  rgb
 * Rearrange structures for consistent alignment within a union.
 * Add an option for clearing SA table.
 *
 * Revision 1.2  1998/07/14 18:05:51  rgb
 * Added #ifdef __KERNEL__ directives to restrict scope of header.
 *
 * Revision 1.1  1998/06/18 21:27:49  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid
 * kernel-build scripts happier in the presence of symlinks
 *
 * Revision 1.4  1998/05/18 21:48:24  rgb
 * Added switch for ungrouping spi's.
 *
 * Revision 1.3  1998/04/23 21:01:50  rgb
 * Added a macro for userspace access to klips kernel debugging switches.
 *
 * Revision 1.2  1998/04/21 21:29:09  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.1  1998/04/09 03:06:09  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:03  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * No changes.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

/*
 * IPSEC <> netlink interface
 */

#define EM_MAXRELSPIS	4		/* at most five chained xforms */
#define EM_MAGIC	0x5377616e	/* "Swan" */

#define EMT_IFADDR	1	/* set enc if addr */
#define EMT_SETSPI	2	/* Set SPI properties */
#define EMT_DELSPI	3	/* Delete an SPI */
#define EMT_GRPSPIS	4	/* Group SPIs (output order)  */
#define EMT_SETEROUTE	5	/* set an extended route */
#define EMT_DELEROUTE	6	/* del an extended route */
#define EMT_TESTROUTE	7	/* try to find route, print to console */
#define EMT_SETDEBUG	8	/* set debug level if active */
#define EMT_UNGRPSPIS	9	/* UnGroup SPIs (output order)  */
#define EMT_CLREROUTE	10	/* clear the extended route table */
#define EMT_CLRSPIS	11	/* clear the spi table */


#define EM_MINLEN	8		/* count!!! */
#define EMT_IFADDR_LEN	12
#define EMT_SETSPI_FLEN	24
#define EMT_GRPSPIS_FLEN 8
#define EMT_SETDEBUG_FLEN 40

#ifdef DEBUG_IPSEC
#define DB_NL_TDBCB	0x0001
#endif

struct encap_msghdr
{
	__u32	em_magic;		/* EM_MAGIC */
	__u16	em_msglen;		/* message length */
	__u8	em_version;		/* for future expansion */
	__u8	em_type;		/* message type */
	union
	{
		__u8	C;		/* Free-text */
		
		struct 
		{
			struct in_addr Dst;
			__u32 Spi;
			struct sockaddr_encap Eaddr;
			struct sockaddr_encap Emask;
		} Ert;

		struct
		{
			struct in_addr Ia;
			__u8	Ifn;
			__u8  xxx[3];	/* makes life a lot easier */
		} Ifa;

		struct
		{
			struct in_addr Dst; /* Destination address */
			__u32 Spi;	/* SPI */
			int If;		/* enc i/f for input */
			int Alg;	/* Algorithm to use */
			__u8 Dat[1];	/* Data */
		} Xfm;
		
		struct
		{
			struct in_addr emr_dst; /* Dest */
			__u32 emr_spi;	/* SPI */
			struct tdb * emr_tdb; /* used internally! */
			
		} Rel[EM_MAXRELSPIS];
		
#ifdef DEBUG_IPSEC
		struct
		{
			int debug_tunnel;
			int debug_netlink;
			int debug_xform;
			int debug_eroute;
			int debug_spi;
			int debug_radij;
			int debug_esp;
			int debug_ah;
		} Dbg;
#endif DEBUG_IPSEC

	} Eu;
};

#define em_c	Eu.C
#define em_eaddr Eu.Ert.Eaddr
#define em_emask Eu.Ert.Emask
#define em_erdst Eu.Ert.Dst
#define em_erspi Eu.Ert.Spi

#define em_ifa	Eu.Ifa.Ia
#define em_ifn	Eu.Ifa.Ifn

#define em_spi	Eu.Xfm.Spi
#define em_dst	Eu.Xfm.Dst
#define em_if	Eu.Xfm.If
#define em_alg	Eu.Xfm.Alg
#define em_dat	Eu.Xfm.Dat

#define em_rel	Eu.Rel

#ifdef DEBUG_IPSEC
#define em_db_tn Eu.Dbg.debug_tunnel
#define em_db_nl Eu.Dbg.debug_netlink
#define em_db_xf Eu.Dbg.debug_xform
#define em_db_er Eu.Dbg.debug_eroute
#define em_db_sp Eu.Dbg.debug_spi
#define em_db_rj Eu.Dbg.debug_radij
#define em_db_es Eu.Dbg.debug_esp
#define em_db_ah Eu.Dbg.debug_ah
#endif DEBUG_IPSEC

#ifdef __KERNEL__
extern int ipsec_callback(struct sk_buff *skb);
extern void ipsec_print_ip(struct iphdr *ip);

#ifdef DEBUG_IPSEC
extern int debug_netlink;
#endif /* DEBUG_IPSEC */
#endif /* __KERNEL__ */


