/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_radij.c,v 1.8 1998/08/06 17:24:23 rgb Exp $
 *
 * $Log: ipsec_radij.c,v $
 * Revision 1.8  1998/08/06 17:24:23  rgb
 * Fix addrtoa return code bug from stale manpage advice preventing packets
 * from being erouted.
 *
 * Revision 1.7  1998/08/06 07:44:59  rgb
 * Fixed /proc/net/ipsec_eroute subnettoa and addrtoa return value bug that
 * ended up in nothing being printed.
 *
 * Revision 1.6  1998/08/05 22:16:41  rgb
 * Cleanup to prevent cosmetic errors (ie. debug output) from being fatal.
 *
 * Revision 1.5  1998/07/29 20:38:44  rgb
 * Debug and fix subnettoa and addrtoa output.
 *
 * Revision 1.4  1998/07/28 00:02:39  rgb
 * Converting to exclusive use of addrtoa.
 * Fix eroute delete.
 *
 * Revision 1.3  1998/07/14 18:21:26  rgb
 * Add function to clear the eroute table.
 *
 * Revision 1.2  1998/06/23 02:59:14  rgb
 * Added debugging output to eroute add/delete routines.
 *
 * Revision 1.9  1998/06/18 21:29:06  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid kernel
 * build scripts happier in presence of symbolic links
 *
 * Revision 1.8  1998/06/05 02:32:26  rgb
 * Fix spi ntoh kernel debug output.
 *
 * Revision 1.7  1998/05/25 20:30:37  rgb
 * Remove temporary ipsec_walk, rj_deltree and rj_delnodes functions.
 *
 * Rename ipsec_rj_walker (ipsec_walk) to ipsec_rj_walker_procprint and
 * add ipsec_rj_walker_delete.
 *
 * Revision 1.6  1998/05/21 13:08:57  rgb
 * Rewrote procinfo subroutines to avoid *bad things* when more that 3k of
 * information is available for printout.
 *
 * Revision 1.5  1998/05/18 21:35:55  rgb
 * Clean up output for numerical consistency and readability.  Zero freed
 * eroute memory.
 *
 * Revision 1.4  1998/04/21 21:28:58  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.3  1998/04/14 17:30:39  rgb
 * Fix up compiling errors for radij tree memory reclamation.
 *
 * Revision 1.2  1998/04/12 22:03:23  rgb
 * Updated ESP-3DES-HMAC-MD5-96,
 * 	ESP-DES-HMAC-MD5-96,
 * 	AH-HMAC-MD5-96,
 * 	AH-HMAC-SHA1-96 since Henry started freeswan cvs repository
 * from old standards (RFC182[5-9] to new (as of March 1998) drafts.
 *
 * Fixed eroute references in /proc/net/ipsec*.
 *
 * Started to patch module unloading memory leaks in ipsec_netlink and
 * radij tree unloading.
 *
 * Revision 1.1  1998/04/09 03:06:10  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:03  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * No changes.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"

/* XXX */
/* #include "atosubnet.c" */
/* #include <freeswan.h> */
#include "../../../lib/freeswan.h"
	
#ifdef DEBUG_IPSEC
int debug_radij = 0;
#endif

/*
 * Interface between the IPSEC code and the radix (radij) tree code.
 */

/* XXX static */ 
struct radij_node_head *rnh = 0;

int
ipsec_radijinit(void)
{
	maj_keylen = sizeof (struct sockaddr_encap);

	rj_init();
	
	if (rj_inithead((void **)&rnh, 16) == 0) /* 16 is bit offset of sen_type */
		return -1;
	return 0;
}

int
ipsec_radijcleanup(void)
{
	radijcleanup();

	return 0;
}

int
ipsec_cleareroutes(void)
{
	return radijcleartree();
}

struct radij_node *
ipsec_breakroute(struct sockaddr_encap *eaddr, struct sockaddr_encap *emask)
{
	struct radij_node *rn;
	char buf1[32], buf2[32];
	
#ifdef DEBUG_IPSEC
	if (debug_eroute) {
/*		printk("ipsec_breakroute: attempting to delete eroute for %u.%u.%u.%u/%u.%u.%u.%u->%u.%u.%u.%u/%u.%u.%u.%u\n\n",
		       O4((eaddr->sen_ip_src.s_addr)),
		       O3((eaddr->sen_ip_src.s_addr)),
		       O2((eaddr->sen_ip_src.s_addr)),
		       O1((eaddr->sen_ip_src.s_addr)),
		       O4((emask->sen_ip_src.s_addr)),
		       O3((emask->sen_ip_src.s_addr)),
		       O2((emask->sen_ip_src.s_addr)),
		       O1((emask->sen_ip_src.s_addr)),
		       O4((eaddr->sen_ip_dst.s_addr)),
		       O3((eaddr->sen_ip_dst.s_addr)),
		       O2((eaddr->sen_ip_dst.s_addr)),
		       O1((eaddr->sen_ip_dst.s_addr)),
		       O4((emask->sen_ip_dst.s_addr)),
		       O3((emask->sen_ip_dst.s_addr)),
		       O2((emask->sen_ip_dst.s_addr)),
		       O1((emask->sen_ip_dst.s_addr))); */
		if(!(subnettoa(eaddr->sen_ip_src, emask->sen_ip_src, 0, buf1, sizeof(buf1)))) {
			printk("ipsec_makeroute: subnetta failed on 0x%0x/0x%0x\n",
			       eaddr->sen_ip_src.s_addr, emask->sen_ip_src.s_addr);
			buf1[0] = '\0';
		}
		if(!(subnettoa(eaddr->sen_ip_dst, emask->sen_ip_dst, 0, buf2, sizeof(buf2)))) {
			printk("ipsec_makeroute: subnetta failed on 0x%0x/0x%0x\n",
			       eaddr->sen_ip_dst.s_addr, emask->sen_ip_dst.s_addr);
			buf2[0] = '\0';
		}
		if(buf1[0] && buf2[0]) {
			printk("ipsec_makeroute: attempting to delete eroute for %s->%s\n",
			       buf1, buf2);
		} else {
			printk("ipsec_makeroute: attempting to delete eroute for 0x%x/0x%x->0x%x/0x%x\n",
			       eaddr->sen_ip_src.s_addr, emask->sen_ip_src.s_addr,
			       eaddr->sen_ip_dst.s_addr, emask->sen_ip_dst.s_addr);
		}
	}
#endif

	if ((rn = rj_delete(eaddr, emask, rnh)) == 0)
	{
		printk("ipsec_breakroute: node not found\n");
		return NULL;
	}
	
	if (rn->rj_flags & (RJF_ACTIVE | RJF_ROOT))
		panic ("ipsec_callback RMT_DELEROUTE root or active node\n");
	
	memset((caddr_t)rn, 0, sizeof (struct eroute));
	kfree_s(rn, sizeof(struct eroute));
	
	return rn;
}

struct eroute *
ipsec_makeroute(struct sockaddr_encap *eaddr, struct sockaddr_encap *emask,
	       struct in_addr dst, u_long spi)
{
	struct eroute *retrt;
	struct radij_node *rn;
	char buf1[32], buf2[32], buf3[16];
	size_t len;
	
#ifdef DEBUG_IPSEC
	if (debug_eroute) {
/*		printk("ipsec_makeroute: attempting to insert eroute for %u.%u.%u.%u/%u.%u.%u.%u->%u.%u.%u.%u/%u.%u.%u.%u, dst=%u.%u.%u.%u, spi=0x%lx\n",
		       O4((eaddr->sen_ip_src.s_addr)),
		       O3((eaddr->sen_ip_src.s_addr)),
		       O2((eaddr->sen_ip_src.s_addr)),
		       O1((eaddr->sen_ip_src.s_addr)),
		       O4((emask->sen_ip_src.s_addr)),
		       O3((emask->sen_ip_src.s_addr)),
		       O2((emask->sen_ip_src.s_addr)),
		       O1((emask->sen_ip_src.s_addr)),
		       O4((eaddr->sen_ip_dst.s_addr)),
		       O3((eaddr->sen_ip_dst.s_addr)),
		       O2((eaddr->sen_ip_dst.s_addr)),
		       O1((eaddr->sen_ip_dst.s_addr)),
		       O4((emask->sen_ip_dst.s_addr)),
		       O3((emask->sen_ip_dst.s_addr)),
		       O2((emask->sen_ip_dst.s_addr)),
		       O1((emask->sen_ip_dst.s_addr)),
		       O4((dst.s_addr)),
	               O3((dst.s_addr)),
		       O2((dst.s_addr)),
		       O1((dst.s_addr)),
		       ntohl(spi)); */
		if(!(len=subnettoa(eaddr->sen_ip_src, emask->sen_ip_src, 0, buf1, sizeof(buf1)))) {
			printk("ipsec_makeroute: subnettoa failed with src=0x%x/0x%0x, len=%d.\n",
			       eaddr->sen_ip_src.s_addr, emask->sen_ip_src.s_addr, len);
			buf1[0] = '\0';
		}
		if(!(len=subnettoa(eaddr->sen_ip_dst, emask->sen_ip_dst, 0, buf2, sizeof(buf2)))) {
			printk("ipsec_makeroute: subnettoa failed with dst=0x%x/0x%0x, len=%d.\n",
			       eaddr->sen_ip_dst.s_addr, emask->sen_ip_dst.s_addr, len);
			buf2[0] = '\0';
		}
		if(!(len=addrtoa(dst, 0, buf3, sizeof(buf3)))) {
			printk("ipsec_makeroute: addrtoa failed with dst=0x%x, len=%d.\n",
			       dst.s_addr, len);
			buf3[0] = '\0';
		}
		if(buf1 && buf2 && buf3) {
			printk("ipsec_makeroute: attempting to insert eroute for %s->%s, dst=%s, spi=0x%lx\n",
			       buf1, buf2,
			       buf3,
			       ntohl(spi));
		} else {
			printk("ipsec_makeroute: attempting to insert eroute for 0x%0x/0x%0x->0x%0x/0x%0x, dst=0x%0x, spi=0x%lx\n",
			       eaddr->sen_ip_src.s_addr, emask->sen_ip_src.s_addr,
			       eaddr->sen_ip_dst.s_addr, emask->sen_ip_dst.s_addr,
			       dst.s_addr, 
			       ntohl(spi));
		}
	}
#endif

	retrt = (struct eroute *)kmalloc(sizeof (struct eroute), GFP_ATOMIC);
	if (retrt == NULL) {
		printk("ipsec_makeroute: not able to allocate kernel memory");
		return NULL;
	}
	memset((caddr_t)retrt, 0, sizeof (struct eroute));

	retrt->er_eaddr = *eaddr;
	retrt->er_emask = *emask;
	retrt->er_dst = dst;
	retrt->er_spi = spi;
	
	rd_key((&(retrt->er_rjt))) = &(retrt->er_eaddr);
	
	rn = rj_addroute(&(retrt->er_eaddr), &(retrt->er_emask), 
			 rnh, retrt->er_rjt.rd_nodes);
	
	if (rn == NULL) {
		if(!(len = addrtoa(dst, 0, buf3, sizeof(buf3)))) {
			printk("ipsec_makeroute: addrtoa failed with dst=%lx, len=%d.\n",
			       ntohl(dst.s_addr),
			       len);
			buf3[0] = '\0';
		}
		if(buf3) {
			printk("ipsec_makeroute: rj_addroute not able to insert eroute for dst=%s, spi=0x%lx\n",
			       buf3,
			       ntohl(spi));
		} else {
			printk("ipsec_makeroute: rj_addroute not able to insert eroute for dst=0x%lx, spi=0x%lx\n",
			       ntohl(dst.s_addr),
			       ntohl(spi));
		}
		kfree(retrt); /* XXX -- should we? */
		return NULL;
	}
#ifdef DEBUG_IPSEC
	if (debug_eroute) {
		printk("ipsec_makeroute: succeeded, I think...\n");
	}	
#endif
	return retrt;
}

struct eroute *
ipsec_findroute(struct sockaddr_encap *eaddr)
{
	struct radij_node *rn;
	char buf1[16], buf2[16];
	
#ifdef DEBUG_IPSEC
	if (debug_radij & DB_RJ_FINDROUTE) {
/*		printk("ipsec_findroute: %u.%u.%u.%u->%u.%u.%u.%u\n",
		       O4((eaddr->sen_ip_src.s_addr)),
		       O3((eaddr->sen_ip_src.s_addr)),
		       O2((eaddr->sen_ip_src.s_addr)),
		       O1((eaddr->sen_ip_src.s_addr)),
		       O4((eaddr->sen_ip_dst.s_addr)),
		       O3((eaddr->sen_ip_dst.s_addr)),
		       O2((eaddr->sen_ip_dst.s_addr)),
		       O1((eaddr->sen_ip_dst.s_addr))); */
		if(!(addrtoa(eaddr->sen_ip_src, 0, buf1, sizeof(buf1)))) {
			return NULL;
		}
		if(!(addrtoa(eaddr->sen_ip_dst, 0, buf2, sizeof(buf2)))) {
			return NULL;
		}
		if(buf1[0] && buf2[0]) {
			printk("ipsec_findroute: %s->%s\n",
			       buf1, buf2);
		}
	}
#endif
	rn = rj_match((caddr_t)eaddr, rnh);
	return (struct eroute *)rn;
}
		
int
ipsec_rj_walker_procprint(struct radij_node *rn, void *w0)
{
	struct eroute *ro = (struct eroute *)rn;
	struct rjtentry *rd = (struct rjtentry *)rn;
	struct wsbuf *w = (struct wsbuf *)w0;
	char buf1[32], buf2[32], buf3[16];
	
	struct sockaddr_encap *key, *mask;
	
/*	u_long as, ad, ms, md, ea; */
	
#ifdef DEBUG_IPSEC
	if(debug_radij)
	{
		printk("ipsec_rj_walker_procprint: rn=%p, w0=%p\n", rn, w0);
	}
#endif DEBUG_IPSEC
	if (rn == NULL)
	{
		return 120;
	}
	
	if (rn->rj_b >= 0)
	{
		return 0;
	}
	
	key = rd_key(rd);
	mask = rd_mask(rd);
	
	if ((key == 0) || (mask == 0))
	{
		return 0;
	}
	
/*	as = key->sen_ip_src.s_addr;
	ad = key->sen_ip_dst.s_addr;
	ms = mask->sen_ip_src.s_addr;
	md = mask->sen_ip_dst.s_addr;
	ea = ro->er_dst.s_addr; 
	
	w->len += sprintf(w->buffer + w->len,
			  "(%d.%d.%d.%d/%d.%d.%d.%d -> %d.%d.%d.%d/%d.%d.%d.%d) => \n    (%d.%d.%d.%d, 0x%x)\n\n",
			  O4(as), O3(as), O2(as), O1(as), 
			  O4(ms), O3(ms), O2(ms), O1(ms), 
			  O4(ad), O3(ad), O2(ad), O1(ad), 
			  O4(md), O3(md), O2(md), O1(md), 
			  O4(ea), O3(ea), O2(ea), O1(ea),
			  (u_int)ntohl(ro->er_spi)); */

	if(!subnettoa(key->sen_ip_src, mask->sen_ip_src, 0, buf1, sizeof(buf1))) {
		printk("ipsec_rj_walker_print: subnettoa failed on 0x%0x/0x%0x.\n",
		       key->sen_ip_src.s_addr, mask->sen_ip_src.s_addr);
		return -1;
	}
	if(!subnettoa(key->sen_ip_dst, mask->sen_ip_dst, 0, buf2, sizeof(buf2))) {
		printk("ipsec_rj_walker_print: subnettoa failed on 0x%0x/0x%0x.\n",
		       key->sen_ip_dst.s_addr, mask->sen_ip_dst.s_addr);
		return -1;
	}
	if(!addrtoa(ro->er_dst, 0, buf3, sizeof(buf3))) {
		printk("ipsec_rj_walker_print: addrtoa failed with dst=0x%x.\n",
		       ro->er_dst.s_addr);
		return -1;
	}

	w->len += sprintf(w->buffer + w->len,
			  "( %s -> %s )=>(%s, 0x%x)\n\n",
			  buf1, buf2,
			  buf3,
			  (u_int)ntohl(ro->er_spi));
	w->buffer[w->len] = '\0';
	w->pos = w->begin + w->len;
	if(w->pos < w->offset)
	{
		w->len = 0;
		w->begin = w->pos;
	}
	else if (w->pos > w->offset + w->length)
	{
		w->len = w->prev_len;
		return ENOBUFS;
	}
	w->prev_len = w->len;

	return 0;
}

int
ipsec_rj_walker_delete(struct radij_node *rn, void *w0)
{
/*	struct eroute *ro = (struct eroute *)rn; */
	struct rjtentry *rd = (struct rjtentry *)rn;
/*	struct wsbuf *w = (struct wsbuf *)w0; */
	struct radij_node *rn2;

	char buf1[32] = { 0 }, buf2[32] = { 0 };
	struct sockaddr_encap *key, *mask;

	if (rn == NULL)
	{
		return 120;
	}
	
	key = rd_key(rd);
	mask = rd_mask(rd);
	
#if 0
	if (rn->rj_b >= 0)
	{
		return 0;
	}
#endif
	
#ifdef DEBUG_IPSEC
	if(debug_radij)
	{
		if(!subnettoa(key->sen_ip_src, mask->sen_ip_src, 0, buf1, sizeof(buf1))) {
			printk("ipsec_rj_walker_delete: subnettoa failed on 0x%0x/0x%0x.\n",
			       key->sen_ip_src.s_addr, mask->sen_ip_src.s_addr);
		}
		if(!subnettoa(key->sen_ip_dst, mask->sen_ip_dst, 0, buf2, sizeof(buf2))) {
			printk("ipsec_rj_walker_delete: subnettoa failed on 0x%0x/0x%0x.\n",
			       key->sen_ip_dst.s_addr, mask->sen_ip_dst.s_addr);
		}
		if(buf1[0] && buf2[0]) {
			printk("ipsec_rj_walker_delete: deleting: %s -> %s \n\n",
			       buf1, buf2);
		} else {
			printk("ipsec_rj_walker_delete: deleting: 0x%0x/0x%0x -> 0x%0x/0x%0x \n\n",
			       key->sen_ip_src.s_addr, mask->sen_ip_src.s_addr,
			       key->sen_ip_dst.s_addr, mask->sen_ip_dst.s_addr);
		}
	}
#endif DEBUG_IPSEC

/*	kfree_s(rn, sizeof(struct radij_node)); */
/*	kfree(rn); */
	if((rn2 = rj_delete(key, mask, rnh)) != rn)
		printk("ipsec_rj_walker_delete: tried to delete a different node?!?  This should never happen!\n");

	memset((caddr_t)rn, 0, sizeof (struct eroute));
	kfree_s(rn, sizeof(struct eroute));
	
	return 0;
}
