/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_xform.c,v 1.4 1998/08/12 00:11:31 rgb Exp $
 *
 * $Log: ipsec_xform.c,v $
 * Revision 1.4  1998/08/12 00:11:31  rgb
 * Added new xform functions to the xform table.
 * Fixed minor debug output spelling error.
 *
 * Revision 1.3  1998/07/09 17:45:31  rgb
 * Clarify algorithm not available message.
 *
 * Revision 1.2  1998/06/23 03:00:51  rgb
 * Check for presence of IPIP protocol if it is setup one way (we don't
 * know what has been set up the other way and can only assume it will be
 * symmetrical with the exception of keys).
 *
 * Revision 1.1  1998/06/18 21:27:51  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid
 * kernel-build scripts happier in the presence of symlinks
 *
 * Revision 1.3  1998/06/11 05:54:59  rgb
 * Added transform version string pointer to xformsw initialisations.
 *
 * Revision 1.2  1998/04/21 21:28:57  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.1  1998/04/09 03:06:13  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:02  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.5  1997/06/03 04:24:48  ji
 * Added ESP-3DES-MD5-96
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * Added new transforms.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

#include <linux/config.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/icmp.h>

#include <net/checksum.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <net/netlink.h>
#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"

#include "ipsec_ipe4.h"

#ifdef CONFIG_IPSEC_AH
#include "ipsec_ah.h"
#endif

#ifdef CONFIG_IPSEC_ESP
#include "ipsec_esp.h"
#endif

#ifdef DEBUG_IPSEC
int debug_xform = 0;
#endif

/*
 * Common routines for IPSEC transformations.
 */

struct xformsw xformsw[] = {
{ XF_IP4,		0,		"IPv4 Simple Encapsulation",	ipe4_version,
  ipe4_attach,		ipe4_init,	ipe4_zeroize,
  ipe4_print,		ipe4_room,
  ipe4_input,	ipe4_output, },

#ifdef CONFIG_IPSEC_AH_MD5
{ XF_AHMD5,		XFT_AUTH,	"Keyed MD5 Authentication",	ahmd5_version,
  ahmd5_attach,		ahmd5_init,	ahmd5_zeroize,	
  ahmd5_print,		ahmd5_room,
  ahmd5_input,		ahmd5_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_DES_OLD
{ XF_ESPDESOLD,		XFT_CONF,	"old DES-CBC Encryption",	espdesold_version,
  espdesold_attach,	espdesold_init,	espdesold_zeroize,
  espdesold_print,	espdesold_room,
  espdesold_input,	espdesold_output, },
#endif
#ifdef CONFIG_IPSEC_AH_HMAC_MD5
{ XF_AHHMACMD5,		XFT_AUTH,	"HMAC MD5 Authentication",	ahhmacmd5_version,
  ahhmacmd5_attach,	ahhmacmd5_init,	ahhmacmd5_zeroize,	
  ahhmacmd5_print,	ahhmacmd5_room,
  ahhmacmd5_input,	ahhmacmd5_output, },
#endif
#ifdef CONFIG_IPSEC_AH_HMAC_SHA1
{ XF_AHHMACSHA1,	XFT_AUTH,	"HMAC SHA-1 Authentication",	ahhmacsha1_version,
  ahhmacsha1_attach,	ahhmacsha1_init,	ahhmacsha1_zeroize,	
  ahhmacsha1_print,	ahhmacsha1_room,
  ahhmacsha1_input,	ahhmacsha1_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_DES_MD5
{ XF_ESPDESMD5,		XFT_CONF,	"DES-MD5 Encryption",		espdesmd5_version,
  espdesmd5_attach,	espdesmd5_init,	espdesmd5_zeroize,
  espdesmd5_print,	espdesmd5_room,
  espdesmd5_input,	espdesmd5_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_3DES_MD5
{ XF_ESP3DESMD5,	XFT_CONF,	"3DES-MD5 Encryption",		esp3desmd5_version,
  esp3desmd5_attach,	esp3desmd5_init,	esp3desmd5_zeroize,
  esp3desmd5_print,	esp3desmd5_room,
  esp3desmd5_input,	esp3desmd5_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_3DES_MD5_96
{ XF_ESP3DESMD596,	XFT_CONF,	"3DES-MD5-96 Encryption",	esp3desmd596_version,
  esp3desmd596_attach,	esp3desmd596_init,	esp3desmd596_zeroize,
  esp3desmd596_print,	esp3desmd596_room,
  esp3desmd596_input,	esp3desmd596_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_DES_MD5_96
{ XF_ESPDESMD596,	XFT_CONF,	"DES-MD5-96 Encryption",	espdesmd596_version,
  espdesmd596_attach,	espdesmd596_init,	espdesmd596_zeroize,
  espdesmd596_print,	espdesmd596_room,
  espdesmd596_input,	espdesmd596_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_NULL_MD5_96
{ XF_ESPNULLMD596,	XFT_CONF,	"NULL-MD5-96 ESP *Plaintext*",	espnullmd596_version,
  espnullmd596_attach,	espnullmd596_init,	espnullmd596_zeroize,
  espnullmd596_print,	espnullmd596_room,
  espnullmd596_input,	espnullmd596_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_3DES_SHA1_96
{ XF_ESP3DESSHA196,	XFT_CONF,	"3DES-SHA1-96 Encryption",	esp3dessha196_version,
  esp3dessha196_attach,	esp3dessha196_init,	esp3dessha196_zeroize,
  esp3dessha196_print,	esp3dessha196_room,
  esp3dessha196_input,	esp3dessha196_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_DES_SHA1_96
{ XF_ESPDESSHA196,	XFT_CONF,	"DES-SHA1-96 Encryption",	espdessha196_version,
  espdessha196_attach,	espdessha196_init,	espdessha196_zeroize,
  espdessha196_print,	espdessha196_room,
  espdessha196_input,	espdessha196_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_NULL_SHA1_96
{ XF_ESPNULLSHA196,	XFT_CONF,	"NULL-SHA1-96 ESP *Plaintext*",	espnullsha196_version,
  espnullsha196_attach,	espnullsha196_init,	espnullsha196_zeroize,
  espnullsha196_print,	espnullsha196_room,
  espnullsha196_input,	espnullsha196_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_DES
{ XF_ESPDES,		XFT_CONF,	"DES Encryption",		espdes_version,
  espdes_attach,	espdes_init,	espdes_zeroize,
  espdes_print,		espdes_room,
  espdes_input,		espdes_output, },
#endif
#ifdef CONFIG_IPSEC_ESP_3DES
{ XF_ESP3DES,		XFT_CONF,	"3DES Encryption",		esp3des_version,
  esp3des_attach,	esp3des_init,	esp3des_zeroize,
  esp3des_print,	esp3des_room,
  esp3des_input,	esp3des_output, },
#endif
};

struct xformsw *xformswNXFORMSW = &xformsw[sizeof(xformsw)/sizeof(xformsw[0])];

struct tdb *
gettdb(u_long spi, struct in_addr dst)
{
	int hashval;
	struct tdb *tdbp;
	
	hashval = (spi+dst.s_addr) % TDB_HASHMOD;
	
	for (tdbp = tdbh[hashval]; tdbp; tdbp = tdbp->tdb_hnext)
		if ((tdbp->tdb_spi == spi) && (tdbp->tdb_dst.s_addr == dst.s_addr))
			break;
	
	return tdbp;
}

void
puttdb(struct tdb *tdbp)
{
	int hashval;
	hashval = ((tdbp->tdb_spi + tdbp->tdb_dst.s_addr) % TDB_HASHMOD);
	tdbp->tdb_hnext = tdbh[hashval];
	tdbh[hashval] = tdbp;
}

void
deltdb(struct tdb *tdbp)
{
	int hashval;
	struct tdb *tdbtp;
	hashval = ((tdbp->tdb_spi + tdbp->tdb_dst.s_addr) % TDB_HASHMOD);
	if (tdbp == tdbh[hashval])
	{
		tdbh[hashval] = tdbp->tdb_hnext;
		tdbp->tdb_hnext = NULL;
	}
	else
		for (tdbtp = tdbh[hashval]; tdbtp; tdbtp = tdbtp->tdb_hnext)
			if (tdbtp->tdb_hnext == tdbp)
			{
				tdbtp->tdb_hnext = tdbp->tdb_hnext;
				tdbp->tdb_hnext = NULL;
				break;
			}
}

int
tdb_init(struct tdb *tdbp, struct encap_msghdr *em)
{
	int alg;
	struct xformsw *xsp;
	
	alg = em->em_alg;
	
	for (xsp = xformsw; xsp < xformswNXFORMSW; xsp++)
		if (xsp->xf_type == alg)
		{
#ifdef INET_GET_PROTOCOL
			if((xsp->xf_type == XF_IP4) && (inet_get_protocol(IPPROTO_IPIP) == NULL))
			{
				printk("tdb_init: warning -- IPPROTO_IPIP not loaded, tunnel in reverse direction will fail if requested.\n");
			}
#endif
			
#ifdef DEBUG_IPSEC
			if(debug_netlink)
				printk("tdb_init: calling init routine of %s\n", xsp->xf_name);
#endif
			return (*(xsp->xf_init))(tdbp, xsp, em);
		}
	
#ifdef DEBUG_IPSEC
	if (debug_xform & DB_XF_INIT)
		printk("tdbinit: unregistered algorithm %d requested -- trying to setup spi:0x%lx, addr:%d.%d.%d.%d\n",
		       alg,
		       ntohl(tdbp->tdb_spi),
		       O4(tdbp->tdb_dst.s_addr),
		       O3(tdbp->tdb_dst.s_addr),
		       O2(tdbp->tdb_dst.s_addr),
		       O1(tdbp->tdb_dst.s_addr));
#endif
	return EINVAL;
}

int 
ipsec_tdbcleanup()
{
	int i;
	struct tdb *tdbp, *tdbp2;

	for (i = 0; i < TDB_HASHMOD; i++)
	{
		tdbp = tdbh[i];
		tdbh[i] = NULL;
		for (; tdbp;)
		{
			tdbp2 = tdbp;
			tdbp = tdbp->tdb_hnext;
			if (tdbp2->tdb_xdata)
			{
				tdbp2->tdb_xform->xf_zeroize(tdbp2->tdb_xdata);
				kfree((caddr_t)tdbp2->tdb_xdata);
				tdbp2->tdb_xdata = NULL;
			}
			memset((caddr_t)tdbp2, 0, sizeof(*tdbp2));
			kfree_s(tdbp2, sizeof(*tdbp2));
		}
	}
	return 0;
}
