/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	 Copyright (C) 1996, 1997, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * $Id: ipsec_xform.h,v 1.5 1998/08/12 00:12:30 rgb Exp $
 *
 * $Log: ipsec_xform.h,v $
 * Revision 1.5  1998/08/12 00:12:30  rgb
 * Added macros for new xforms.  Added prototypes for new xforms.
 *
 * Revision 1.4  1998/07/28 00:04:20  rgb
 * Add macro for clearing the SA table.
 *
 * Revision 1.3  1998/07/14 18:06:46  rgb
 * Added #ifdef __KERNEL__ directives to restrict scope of header.
 *
 * Revision 1.2  1998/06/23 03:02:19  rgb
 * Created a prototype for ipsec_tdbcleanup when it was moved from
 * ipsec_init.c.
 *
 * Revision 1.1  1998/06/18 21:27:51  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid
 * kernel-build scripts happier in the presence of symlinks
 *
 * Revision 1.4  1998/06/11 05:55:31  rgb
 * Added transform version string pointer to xformsw structure definition.
 * Added extern declarations for transform version strings.
 *
 * Revision 1.3  1998/05/18 22:02:54  rgb
 * Modify the *_zeroize function prototypes to include one parameter.
 *
 * Revision 1.2  1998/04/21 21:29:08  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.1  1998/04/09 03:06:14  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:06  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.5  1997/06/03 04:24:48  ji
 * Added ESP-3DES-MD5-96
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * Added new transforms.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */

/*
 * Definitions relevant to IPSEC transformations.
 */

#define XF_IP4		1	/* IP inside IP */
#define XF_AHMD5	2	/* AH MD5 */
#define XF_AHSHA	3	/* AH SHA */
#define XF_ESPDESOLD	4	/* old ESP DES-CBC */
#define XF_ESP3DES	5	/* ESP DES3-CBC */
#define XF_AHHMACMD5	6	/* AH-HMAC-MD5 with opt replay prot */
#define XF_AHHMACSHA1	7	/* AH-HMAC-SHA1 with opt replay prot */
#define XF_ESPDESMD5	8	/* DES, HMAC-MD-5 with 128-bits of authentication */
#define XF_ESP3DESMD5	9	/* triple DES, HMAC-MD-5, 128-bits of authentication */
#define	XF_ESP3DESMD596	10	/* triple DES, HMAC-MD-5, 96-bits of authentication */
#define	XF_ESPDESMD596	11	/* DES, HMAC-MD-5 with 96-bits of authentication */
#define	XF_ESPNULLMD596	12	/* NULL, HMAC-MD-5 with 96-bits of authentication */
#define	XF_ESPNULLSHA196	13	/* NULL, HMAC-SHA-1 with 96-bits of authentication */
#define	XF_ESP3DESSHA196	14	/* triple DES, HMAC-SHA-1, 96-bits of authentication */
#define	XF_ESPDESSHA196	15	/* DES, HMAC-SHA-1 with 96-bits of authentication */
#define XF_ESPDES	16	/* ESP DES */
#define XF_CLR		126	/* Clear SA table */
#define XF_DEL		127	/* Delete SA */

#define XFT_AUTH	0x0001
#define XFT_CONF	0x0100

#ifdef DEBUG_IPSEC
#define DB_XF_INIT	0x0001
#endif

#ifdef __KERNEL__
struct tdb				/* tunnel descriptor block */
{
	struct tdb	*tdb_hnext;	/* next in hash chain */
	struct tdb	*tdb_onext;	/* next in output */
	struct tdb	*tdb_inext;	/* next in input (prev!) */
	__u32		tdb_spi;	/* SPI to use (in network order)*/
	struct in_addr	tdb_dst;	/* dest address for this SPI */
	struct ifnet	*tdb_rcvif;	/* related rcv encap interface */
	struct xformsw	*tdb_xform;	/* transformation to use (host order)*/
	caddr_t		tdb_xdata;	/* transformation data (opaque) */
};

#define TDB_HASHMOD	257

struct xformsw
{
	u_short		xf_type;	/* Unique ID of xform */
	u_short		xf_flags;	/* secondary type reall) */
	char		*xf_name;	/* human-readable name */
	char		*xf_cvsid;	/* source code cvs version id */
	int		(*xf_attach)(void);	/* called at config time */
	int		(*xf_init)(struct tdb *tdbp, struct xformsw *xsp, struct encap_msghdr *em);	/* xform initialization */
	int		(*xf_zeroize)(void*);	/* termination */
	int		(*xf_print)(void *, char *);	/* /proc printing */
	int		(*xf_room)(struct tdb *, int, int, int *, int *); /* required head/tail room  */
	struct sk_buff 	*(*xf_input)(struct sk_buff *, struct tdb *);	/* called when packet received */
	int		(*xf_output)(struct sk_buff *, struct tdb *); /* called when packet sent */
};

struct tdb *tdbh[TDB_HASHMOD];
extern struct xformsw xformsw[], *xformswNXFORMSW;

extern struct tdb *gettdb(u_long, struct in_addr);
extern void deltdb(struct tdb *);
extern void puttdb(struct tdb *);
extern int tdb_init(struct tdb *, struct encap_msghdr *);
extern int ipsec_tdbcleanup(void);

extern char ipe4_version[];
extern int ipe4_attach(void);
extern int ipe4_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int ipe4_zeroize(void*);
extern int ipe4_print(void *, char *);
extern int ipe4_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *ipe4_input(struct sk_buff *, struct tdb *);
extern int ipe4_output(struct sk_buff *, struct tdb *);


#ifdef CONFIG_IPSEC_AH_MD5
extern char ahmd5_version[];
extern int ahmd5_attach(void);
extern int ahmd5_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int ahmd5_zeroize(void*);
extern int ahmd5_print(void *, char *);
extern int ahmd5_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *ahmd5_input(struct sk_buff *, struct tdb *);
extern int ahmd5_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_AH_HMAC_MD5
extern char ahhmacmd5_version[];
extern int ahhmacmd5_attach(void);
extern int ahhmacmd5_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int ahhmacmd5_zeroize(void*);
extern int ahhmacmd5_print(void *, char *);
extern int ahhmacmd5_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *ahhmacmd5_input(struct sk_buff *, struct tdb *);
extern int ahhmacmd5_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_AH_HMAC_SHA1
extern char ahhmacsha1_version[];
extern int ahhmacsha1_attach(void);
extern int ahhmacsha1_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int ahhmacsha1_zeroize(void*);
extern int ahhmacsha1_print(void *, char *);
extern int ahhmacsha1_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *ahhmacsha1_input(struct sk_buff *, struct tdb *);
extern int ahhmacsha1_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_DES_OLD
extern char espdesold_version[];
extern int espdesold_attach(void);
extern int espdesold_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espdesold_zeroize(void*);
extern int espdesold_print(void *, char *);
extern int espdesold_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espdesold_input(struct sk_buff *, struct tdb *);
extern int espdesold_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef CONFIG_IPSEC_ESP_DES_MD5
extern char espdesmd5_version[];
extern int espdesmd5_attach(void);
extern int espdesmd5_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espdesmd5_zeroize(void*);
extern int espdesmd5_print(void *, char *);
extern int espdesmd5_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espdesmd5_input(struct sk_buff *, struct tdb *);
extern int espdesmd5_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef CONFIG_IPSEC_ESP_3DES_MD5
extern char esp3desmd5_version[];
extern int esp3desmd5_attach(void);
extern int esp3desmd5_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int esp3desmd5_zeroize(void*);
extern int esp3desmd5_print(void *, char *);
extern int esp3desmd5_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *esp3desmd5_input(struct sk_buff *, struct tdb *);
extern int esp3desmd5_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_3DES_MD5_96
extern char esp3desmd596_version[];
extern int esp3desmd596_attach(void);
extern int esp3desmd596_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int esp3desmd596_zeroize(void*);
extern int esp3desmd596_print(void *, char *);
extern int esp3desmd596_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *esp3desmd596_input(struct sk_buff *, struct tdb *);
extern int esp3desmd596_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef CONFIG_IPSEC_ESP_3DES_SHA1_96
extern char esp3dessha196_version[];
extern int esp3dessha196_attach(void);
extern int esp3dessha196_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int esp3dessha196_zeroize(void*);
extern int esp3dessha196_print(void *, char *);
extern int esp3dessha196_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *esp3dessha196_input(struct sk_buff *, struct tdb *);
extern int esp3dessha196_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef CONFIG_IPSEC_ESP_DES_MD5_96
extern char espdesmd596_version[];
extern int espdesmd596_attach(void);
extern int espdesmd596_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espdesmd596_zeroize(void*);
extern int espdesmd596_print(void *, char *);
extern int espdesmd596_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espdesmd596_input(struct sk_buff *, struct tdb *);
extern int espdesmd596_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_DES_SHA1_96
extern char espdessha196_version[];
extern int espdessha196_attach(void);
extern int espdessha196_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espdessha196_zeroize(void*);
extern int espdessha196_print(void *, char *);
extern int espdessha196_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espdessha196_input(struct sk_buff *, struct tdb *);
extern int espdessha196_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_NULL_MD5_96
extern char espnullmd596_version[];
extern int espnullmd596_attach(void);
extern int espnullmd596_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espnullmd596_zeroize(void*);
extern int espnullmd596_print(void *, char *);
extern int espnullmd596_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espnullmd596_input(struct sk_buff *, struct tdb *);
extern int espnullmd596_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_NULL_SHA1_96
extern char espnullsha196_version[];
extern int espnullsha196_attach(void);
extern int espnullsha196_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espnullsha196_zeroize(void*);
extern int espnullsha196_print(void *, char *);
extern int espnullsha196_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espnullsha196_input(struct sk_buff *, struct tdb *);
extern int espnullsha196_output(struct sk_buff *, struct tdb *);
#endif

#ifdef CONFIG_IPSEC_ESP_DES
extern char espdes_version[];
extern int espdes_attach(void);
extern int espdes_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int espdes_zeroize(void*);
extern int espdes_print(void *, char *);
extern int espdes_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *espdes_input(struct sk_buff *, struct tdb *);
extern int espdes_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef CONFIG_IPSEC_ESP_3DES
extern char esp3des_version[];
extern int esp3des_attach(void);
extern int esp3des_init(struct tdb *, struct xformsw *, struct encap_msghdr *);
extern int esp3des_zeroize(void*);
extern int esp3des_print(void *, char *);
extern int esp3des_room(struct tdb *, int, int, int *, int *);
extern struct sk_buff *esp3des_input(struct sk_buff *, struct tdb *);
extern int esp3des_output(struct sk_buff *, struct tdb *);
#endif 

#ifdef DEBUG_IPSEC
extern int debug_xform;
#endif
#endif /* __KERNEL__ */
