/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	    Copyright (C) 1996, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

char eroute_c_version[] = "$Id: eroute.c,v 1.9 1998/08/18 17:18:13 rgb Exp $";

/*
 * $Log: eroute.c,v $
 * Revision 1.9  1998/08/18 17:18:13  rgb
 * Delete old commented out cruft.
 * Reduce destination and source default subnet to warning, not fatal.
 *
 * Revision 1.8  1998/08/05 22:24:45  rgb
 * Change includes to accomodate RH5.x
 *
 * Revision 1.7  1998/07/29 20:49:08  rgb
 * Change to use 0x-prefixed hexadecimal for spi's.
 *
 * Revision 1.6  1998/07/28 00:14:24  rgb
 * Convert from positional parameters to long options.
 * Add --clean option.
 * Add hostname lookup support.
 *
 * Revision 1.5  1998/07/14 18:13:28  rgb
 * Restructured for better argument checking.
 * Added command to clear the eroute table.
 *
 * Revision 1.4  1998/07/09 18:14:10  rgb
 * Added error checking to IP's and keys.
 * Made most error messages more specific rather than spamming usage text.
 * Added more descriptive kernel error return codes and messages.
 * Converted all spi translations to unsigned.
 * Removed all invocations of perror.
 *
 * Revision 1.3  1998/05/27 18:48:19  rgb
 * Adding --help and --version directives.
 *
 * Revision 1.2  1998/04/13 03:15:29  rgb
 * Commands are now distinguishable from arguments when invoking usage.
 *
 * Revision 1.1.1.1  1998/04/08 05:35:10  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.3  1996/11/20 14:51:32  ji
 * Fixed problems with #include paths.
 * Changed (incorrect) references to ipsp into ipsec.
 *
 * Revision 0.2  1996/11/08 15:45:24  ji
 * First limited release.
 *
 *
 */

#include <sys/types.h>
#include <linux/types.h> /* new */
#include <string.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>


#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_netlink.h"


#include <stdio.h>
#include <getopt.h>

/* XXX */
/* #include "atosubnet.c" */
#include <freeswan.h>

char *program_name;
extern char *optarg;
extern int optind, opterr, optopt;
		
void
usage(char* arg)
{
  fprintf(stderr, "usage: %s --add --src <src>/<srcmaskbits>|<srcmask> --dst <dst>/<dstmaskbits>|<dstmask> --edst <edst> --spi <spi>\n", arg);
  fprintf(stderr, "       %s --del --src <src>/<srcmaskbits>|<srcmask> --dst <dst>/<dstmaskbits>|<dstmask>\n", arg);
  fprintf(stderr, "       %s --clear\n", arg);
  fprintf(stderr, "       %s --help\n", arg);
  fprintf(stderr, "       %s --version\n", arg);
  exit(1);
}

int
resolve_ip(char *arg, struct in_addr *ip)
{
	struct hostent *hostp;
	struct netent *netp;
	
	if (!strcmp(arg, "default")) {/* default is really 0.0.0.0 */
		ip->s_addr = htonl(INADDR_ANY);
		return(1);
	}
	if (!(strcmp(arg, "255.255.255.255"))) {
		/* inet_addr() returns -1 for this, as it should, but that means an error */
		ip->s_addr = htonl(INADDR_BROADCAST);
		return (1);
	}
	if ((netp = getnetbyname(arg)) != (struct netent *)NULL) {
		ip->s_addr = htonl(netp->n_net);
		return(1);
	}
	if ((hostp = gethostbyname(arg)) == (struct hostent *)NULL) {
		errno = h_errno;
		return(-1);
	}
	memcpy((char *) ip, (char *) hostp->h_addr_list[0], hostp->h_length);
	return(0);
}

static struct option const longopts[] =
{
	{"dst", 1, 0, 'D'},
	{"src", 1, 0, 'S'},
	{"add", 0, 0, 'a'},
	{"clear", 0, 0, 'c'},
	{"del", 0, 0, 'd'},
	{"edst", 1, 0, 'e'},
	{"help", 0, 0, 'h'},
	{"spi", 1, 0, 's'},
	{"version", 0, 0, 'v'},
	{0, 0, 0, 0}
};

main(argc, argv)
int argc;
char **argv;
{
	int fd;
	struct encap_msghdr emh;
	char *endptr;
	int ret;
	int c, previous = -1;

	program_name = argv[0];
	memset(&emh, 0, sizeof(emh));

	while((c = getopt_long(argc, argv, "acdD:e:hs:S:v", longopts, 0)) != EOF)
	{
		switch(c) {
		case 'a':
			if(emh.em_type) {
				fprintf(stderr, "%s: Only one of '--add', '--clear', or '--del' options permitted.\n",
					program_name);
				exit(1);
			}
			emh.em_type = EMT_SETEROUTE;
			break;
		case 'c':
			if(emh.em_type) {
				fprintf(stderr, "%s: Only one of '--add', '--clear', or '--del' options permitted.\n",
					program_name);
				exit(1);
			}
			emh.em_type = EMT_CLREROUTE;
			break;
		case 'd':
			if(emh.em_type) {
				fprintf(stderr, "%s: Only one of '--add', '--clear', or '--del' options permitted.\n",
					program_name);
				exit(1);
			}
			emh.em_type = EMT_DELEROUTE;
			break;
		case 'e':
			if(ret = resolve_ip(optarg, &(emh.em_erdst)) != 0) {
				if(ret == 1) {
					fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
						program_name, optarg);
					exit (1);
				} else {
					fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
						program_name, optarg);
					exit (1);
				}
			}
			break;
		case 'h':
		case '?':
			usage(program_name);
			exit(1);
		case 's':
			emh.em_erspi = htonl(strtoul(optarg, &endptr, 0));
			if(!(endptr == optarg + strlen(optarg))) {
				fprintf(stderr, "%s: Invalid character in SPI parameter: %s\n",
					program_name, optarg);
				exit (1);
			}
			break;
		case 'v':
			printf("%s, %s\n", program_name, eroute_c_version);
			exit(1);
		case 'D':
			if (atosubnet(optarg, 0, &emh.em_eaddr.sen_ip_dst, &emh.em_emask.sen_ip_dst)) {
				fprintf(stderr, "%s: Invalid formation of destination network/mask: %s\n",
					program_name, optarg);
				exit (1);
			}
			break;
		case 'S':
			if (atosubnet(optarg, 0, &emh.em_eaddr.sen_ip_src, &emh.em_emask.sen_ip_src)) {
				fprintf(stderr, "%s: Invalid formation of source network/mask: %s\n",
					program_name, optarg);
				exit (1);
			}
			break;
		default:
		}
		previous = c;
	}

#ifdef DEBUG
	printf("%s: DEBUG: argc=%d\n", program_name, argc);
#endif /* DEBUG */

	/* Sanity checks */

	switch(emh.em_type) {
	case EMT_SETEROUTE:
		if(!emh.em_erdst.s_addr) {
			fprintf(stderr, "%s: add option must have '--edst' option specified.\n",
				program_name);
			exit(1);
		}
		if(!emh.em_erspi) {
			fprintf(stderr, "%s: add option must have '--spi' option specified.\n",
				program_name);
			exit(1);
		}
	case EMT_DELEROUTE:
		if(!emh.em_eaddr.sen_ip_src.s_addr || !emh.em_emask.sen_ip_src.s_addr) {
			fprintf(stderr, "%s: warning -- %s option '--src' is 0.0.0.0.\n",
				program_name, (emh.em_type == EMT_SETEROUTE) ? "add" : "del");
/*			exit(1); */
		}
		if(!emh.em_eaddr.sen_ip_dst.s_addr || !emh.em_emask.sen_ip_dst.s_addr) {
			fprintf(stderr, "%s: warning -- %s option '--dst' is 0.0.0.0.\n",
				program_name, (emh.em_type == EMT_SETEROUTE) ? "add" : "del");
/*			exit(1); */
		}
	case EMT_CLREROUTE:
		emh.em_magic = EM_MAGIC;
		emh.em_msglen= sizeof emh;
		emh.em_version = 0;
		emh.em_eaddr.sen_len = emh.em_emask.sen_len = sizeof (struct sockaddr_encap);
		emh.em_eaddr.sen_family = emh.em_emask.sen_family = 26;
		emh.em_eaddr.sen_type = SENT_IP4;
		emh.em_emask.sen_type = 255;
		break;
	default:
		fprintf(stderr, "%s: exactly one of '--add', '--del' or '--clear' options must be specified.\n"
			"Try %s --help' for usage information.\n",
			program_name, program_name);
		exit(1);
	}

	
#if 0
	switch(emh.em_type) {
	case EMT_SETEROUTE:
/*		emh.em_erdst.s_addr = inet_addr(argv[6]); */
		if(ret = resolve_ip(argv[4], &(emh.em_erdst)) != 0) {
			if(ret == 1) {
				fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
					argv[0], argv[4]);
				exit (1);
			} else {
				fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
					argv[0], argv[4]);
				exit (1);
			}
		}

#ifdef DEBUG
	printf("%s: DEBUG: resolver found %s=%s\n", argv[0], argv[4], inet_ntoa(emh.em_erdst));
#endif /* DEBUG */

		emh.em_erspi = htonl(strtoul(argv[5], &endptr, 16));
		if(!(endptr == argv[5] + strlen(argv[5]))) {
			fprintf(stderr, "%s: Invalid character in SPI parameter: %s\n",
				argv[0], argv[5]);
			exit (1);
		}

#ifdef DEBUG
	printf("%s: DEBUG: set ok\n", argv[0]);
#endif /* DEBUG */

	case EMT_DELEROUTE:
/*		emh.em_eaddr.sen_ip_src.s_addr = inet_addr(argv[2]);
		emh.em_eaddr.sen_ip_dst.s_addr = inet_addr(argv[4]);
		emh.em_emask.sen_ip_src.s_addr = inet_addr(argv[3]);
		emh.em_emask.sen_ip_dst.s_addr = inet_addr(argv[5]); */

#ifdef DEBUG
	printf("%s: DEBUG: del ok\n", argv[0]);
#endif /* DEBUG */

		if (atosubnet(argv[2], 0, &emh.em_eaddr.sen_ip_src, &emh.em_emask.sen_ip_src)) {
			fprintf(stderr, "%s: Invalid formation of source network/mask: %s\n",
				argv[0], argv[2]);
			exit (1);
		}
		if (atosubnet(argv[3], 0, &emh.em_eaddr.sen_ip_dst, &emh.em_emask.sen_ip_dst)) {
			fprintf(stderr, "%s: Invalid formation of destination network/mask: %s\n",
				argv[0], argv[3]);
			exit (1);
		}

#ifdef DEBUG
	printf("%s: DEBUG: del ok\n", argv[0]);
#endif /* DEBUG */

	case EMT_CLREROUTE:
		break;
	default:
		usage(argv[0]);
	}
#endif /* 0 */

	fd = open("/dev/ipsec", O_RDWR);
	if (fd < 0) {
		printf("%s: Could not open /dev/ipsec -- ", program_name);
		switch(errno) {
		case ENOENT:
			printf("device does not exist.  See FreeS/WAN installation procedure.\n");
		case EACCES:
			printf("access denied.  ");
			if(getuid() == 0) {
				printf("Check permissions.  Should be 600.\n");
			} else {
				printf("You must be root to open this file.\n");
			}
			break;
		case EUNATCH:
			printf("Netlink not enabled OR KLIPS not loaded.\n");
			break;
		default:
			printf("Unknown file open error %d\n", errno);
		}
		exit(1);
	}

#ifdef DEBUG
	printf("%s: DEBUG: open ok\n", program_name);
#endif /* DEBUG */

	if (write(fd, &emh, sizeof emh) != sizeof emh) {
		printf("%s: Had trouble writing to /dev/ipsec -- ", program_name);
		switch(errno) {
		case EINVAL:
			printf("Invalid argument, check kernel log messages for specifics.\n");
			break;
		default:
			printf("Unknown file write error %d\n", errno);
		}
		close(fd);
		exit(1);
	}
	close(fd);

#ifdef DEBUG
	printf("%s: DEBUG: write ok\n", program_name);
#endif /* DEBUG */

	exit(0);
}

	
	
