/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	    Copyright (C) 1996, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

char klipsdebug_c_version[] = "$Id: klipsdebug.c,v 1.9 1998/08/28 03:13:05 rgb Exp $";

/*
 * $Log: klipsdebug.c,v $
 * Revision 1.9  1998/08/28 03:13:05  rgb
 * Tidy up old cruft.
 *
 * Revision 1.8  1998/08/05 22:24:45  rgb
 * Change includes to accomodate RH5.x
 *
 * Revision 1.7  1998/07/29 21:36:37  rgb
 * Converted to long option names.
 *
 * Revision 1.6  1998/07/14 18:23:11  rgb
 * Remove unused skbuff header.
 *
 * Revision 1.5  1998/07/09 18:14:10  rgb
 * Added error checking to IP's and keys.
 * Made most error messages more specific rather than spamming usage text.
 * Added more descriptive kernel error return codes and messages.
 * Converted all spi translations to unsigned.
 * Removed all invocations of perror.
 *
 * Revision 1.4  1998/05/27 18:48:21  rgb
 * Adding --help and --version directives.
 *
 * Revision 1.3  1998/05/18 21:19:09  rgb
 * Added options for finer control of debugging switches.
 *
 * Revision 1.2  1998/05/12 02:26:27  rgb
 * Fixed compile errors with IPSEC_DEBUG shut off in the kernel config.
 *
 * Revision 1.1  1998/04/23 21:07:34  rgb
 * Added a userspace utility to change klips kernelspace debug switches.
 *
 * Revision 1.1.1.1  1998/04/08 05:35:09  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.3  1996/11/20 14:51:32  ji
 * Fixed problems with #include paths.
 * Changed (incorrect) references to ipsp into ipsec.
 *
 * Revision 0.2  1996/11/08 15:46:29  ji
 * First limited release.
 *
 *
 */

#include <sys/types.h>
#include <linux/types.h> /* new */
#include <string.h>
#include <errno.h>

#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>



#include <unistd.h>
#include "radij.h"
#include "ipsec_encap.h"
#ifndef DEBUG_IPSEC
#define DEBUG_IPSEC
#endif DEBUG_IPSEC
#include "ipsec_netlink.h"

#include <stdio.h>
#include <getopt.h>

u_long bigbuf[1024];
char *program_name;

inline int x2i(char *s)
{
	char ss[3];
	char *endptr;
	long n;
	ss[0] = s[0];
	ss[1] = s[1];
	ss[2] = 0;

	n = strtoul(ss, &endptr, 16);
	if(!(endptr == ss + strlen(ss)))
	{
		fprintf(stderr, "%s: Invalid character in numerical input stream: %s\n",
			program_name, ss);
		exit (1);
	}
	return n;
}
	
void
usage(char * arg)
{
	fprintf(stderr, "usage: %s {--set|--clear} {tunnel|netlink|xform|eroute|spi|radij|esp|ah}\n", arg);
	fprintf(stderr, "       %s {--all|--none}\n", arg);
	fprintf(stderr, "       %s --help\n", arg);
	fprintf(stderr, "       %s --version\n", arg);
	exit(1);
}

static struct option const longopts[] =
{
	{"set", 1, 0, 's'},
	{"clear", 1, 0, 'c'},
	{"all", 0, 0, 'a'},
	{"none", 0, 0, 'n'},
	{"help", 0, 0, 'h'},
	{"version", 0, 0, 'v'},
	{0, 0, 0, 0}
};

main(argc, argv)
int argc;
char **argv;
{
	int fd;
	unsigned char action = 0;
	int i, dlen, *debug;
	int c, previous = -1;
	
	struct encap_msghdr *em = (struct encap_msghdr *)bigbuf;
	debug = (int *)&em->em_db_tn;

	program_name = argv[0];

	while((c = getopt_long(argc, argv, "s:c:anhv", longopts, 0)) != EOF)
	{
		switch(c) {
		case 's':
			if(action) {
				fprintf(stderr, "%s: Only one of '--set', '--clear', '--all' or '--none' options permitted.\n",
					program_name);
				exit(1);
			}
			action = 's';
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++) {
				*(debug++) = 0;
			}
			if(strcmp(optarg, "tunnel") == 0) {
				em->em_db_tn = -1L;
			} else if(strcmp(optarg, "netlink") == 0) {
				em->em_db_nl = -1L;
			} else if(strcmp(optarg, "xform") == 0) {
				em->em_db_xf = -1L;
			} else if(strcmp(optarg, "eroute") == 0) {
				em->em_db_er = -1L;
			} else if(strcmp(optarg, "spi") == 0) {
				em->em_db_sp = -1L;
			} else if(strcmp(optarg, "radij") == 0) {
				em->em_db_rj = -1L;
			} else if(strcmp(optarg, "esp") == 0) {
				em->em_db_es = -1L;
			} else if(strcmp(optarg, "ah") == 0) {
				em->em_db_ah = -1L;
			} else {
				usage(program_name);
			}
			em->em_db_nl |= 1 << (sizeof(em->em_db_nl) * 8 -1);
			break;
		case 'c':
			if(action) {
				fprintf(stderr, "%s: Only one of '--set', '--clear', '--all' or '--none' options permitted.\n",
					program_name);
				exit(1);
			}
			action = 'c';
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++) {
				*(debug++) = -1;
			}
			if(strcmp(optarg, "tunnel") == 0){
				em->em_db_tn = 0;
			}else if(strcmp(optarg, "netlink") == 0){
				em->em_db_nl = 0;
			}else if(strcmp(optarg, "xform") == 0){
				em->em_db_xf = 0;
			}else if(strcmp(optarg, "eroute") == 0){
				em->em_db_er = 0;
			}else if(strcmp(optarg, "spi") == 0){
				em->em_db_sp = 0;
			}else if(strcmp(optarg, "radij") == 0){
				em->em_db_rj = 0;
			}else if(strcmp(optarg, "esp") == 0){
				em->em_db_es = 0;
			}else if(strcmp(optarg, "ah") == 0){
				em->em_db_ah = 0;
			}else{
				usage(program_name);
			}
			em->em_db_nl &= ~(1 << (sizeof(em->em_db_nl) * 8 -1));
			break;
		case 'a':
			if(action) {
				fprintf(stderr, "%s: Only one of '--set', '--clear', '--all' or '--none' options permitted.\n",
					program_name);
				exit(1);
			}
			action = 'a';
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++) {
				*(debug++) = -1;
			}
			break;
		case 'n':
			if(action) {
				fprintf(stderr, "%s: Only one of '--set', '--clear', '--all' or '--none' options permitted.\n",
					program_name);
				exit(1);
			}
			action = 'n';
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++) {
				*(debug++) = 0;
			}
			break;
		case 'h':
		case '?':
			usage(program_name);
			exit(1);
		case 'v':
			printf("%s, %s\n", program_name, klipsdebug_c_version);
			exit(1);
		default:
		}
		previous = c;
	}

#if 0
	if (argc == 1)
		usage(argv[0]);

        if(strcmp(argv[1], "--help") == 0)
        {
                usage(argv[0]);
        }

        if(strcmp(argv[1], "--version") == 0)
        {
                printf("%s, %s\n", argv[0], klipsdebug_c_version);
                exit(1);
        }

	if (argc < 3)
		usage(argv[0]);
#endif

	if(!action) {
		usage(program_name);
	}
 	em->em_magic = EM_MAGIC;
	em->em_version = 0;
	em->em_type = EMT_SETDEBUG;
	em->em_msglen = EMT_SETDEBUG_FLEN;

#if 0
	if(strcmp(argv[1], "set") == 0)
	{
		debug = (int *)&em->em_db_tn;
		for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++)
		{
			*(debug++) = 0;
		}
		if(strcmp(argv[2], "all") == 0)
		{
			debug = (int *)&em->em_db_tn;
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++)
			{
				*(debug++) = -1;
			}
		}
		else if(strcmp(argv[2], "tunnel") == 0)
		{
			em->em_db_tn = -1L;
		}
		else if(strcmp(argv[2], "netlink") == 0)
		{
			em->em_db_nl = -1L;
		}
		else if(strcmp(argv[2], "xform") == 0)
		{
			em->em_db_xf = -1L;
		}
		else if(strcmp(argv[2], "eroute") == 0)
		{
			em->em_db_er = -1L;
		}
		else if(strcmp(argv[2], "spi") == 0)
		{
			em->em_db_sp = -1L;
		}
		else if(strcmp(argv[2], "radij") == 0)
		{
			em->em_db_rj = -1L;
		}
		else if(strcmp(argv[2], "esp") == 0)
		{
			em->em_db_es = -1L;
		}
		else if(strcmp(argv[2], "ah") == 0)
		{
			em->em_db_ah = -1L;
		}
		else
		{
			usage(argv[0]);
		}
		em->em_db_nl |= 1 << (sizeof(em->em_db_nl) * 8 -1);
	}
	else if(strcmp(argv[1], "unset") == 0)
	{
		debug = (int *)&em->em_db_tn;
		for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++)
		{
			*(debug++) = -1;
		}
		if(strcmp(argv[2], "all") == 0)
		{
			debug = (int *)&em->em_db_tn;
			for (i = 0; i < (sizeof(em->Eu.Dbg)/sizeof(int)); i++)
			{
				*(debug++) = 0;
			}
		}
		else if(strcmp(argv[2], "tunnel") == 0)
		{
			em->em_db_tn = 0;
		}
		else if(strcmp(argv[2], "netlink") == 0)
		{
			em->em_db_nl = 0;
		}
		else if(strcmp(argv[2], "xform") == 0)
		{
			em->em_db_xf = 0;
		}
		else if(strcmp(argv[2], "eroute") == 0)
		{
			em->em_db_er = 0;
		}
		else if(strcmp(argv[2], "spi") == 0)
		{
			em->em_db_sp = 0;
		}
		else if(strcmp(argv[2], "radij") == 0)
		{
			em->em_db_rj = 0;
		}
		else if(strcmp(argv[2], "esp") == 0)
		{
			em->em_db_es = 0;
		}
		else if(strcmp(argv[2], "ah") == 0)
		{
			em->em_db_ah = 0;
		}
		else
		{
			usage(argv[0]);
		}
		em->em_db_nl &= ~(1 << (sizeof(em->em_db_nl) * 8 -1));
	}
	else
	{
		usage(argv[0]);
	}
#endif

	fd = open("/dev/ipsec", 2);
	if (fd < 0)
	{
		fprintf(stderr, "%s: Could not open /dev/ipsec -- ", argv[0]);
		switch(errno){
		case ENOENT:
			fprintf(stderr, "device does not exist.  See FreeS/WAN installation procedure.\n");
		case EACCES:
			fprintf(stderr, "access denied.  ");
			if(getuid() == 0){
				fprintf(stderr, "Check permissions.  Should be 600.\n");
			}else{
				fprintf(stderr, "You must be root to open this file.\n");
			}
			break;
		case EUNATCH:
			fprintf(stderr, "Netlink not enabled OR KLIPS not loaded.\n");
			break;
		default:
			fprintf(stderr, "Unknown file open error %d\n", errno);
		}
		exit(1);
	}
	if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen){
		fprintf(stderr, "%s: Had trouble writing to /dev/ipsec -- ", argv[0]);
		switch(errno){
		case EINVAL:
			fprintf(stderr, "Invalid argument, check kernel log messages for specifics.\n");
			break;
		default:
			fprintf(stderr, "Unknown file write error %d\n", errno);
		}
		close(fd);
		exit(1);
	}
	close(fd);
	exit(0);
}
