/*
 *
 *			   IPSEC for Linux
 *		         Preliminary Release
 * 
 *	    Copyright (C) 1996, John Ioannidis <ji@hol.gr>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

char spigrp_c_version[] = "$Id: spigrp.c,v 1.8 1998/08/05 22:24:45 rgb Exp $";

/*
 * $Log: spigrp.c,v $
 * Revision 1.8  1998/08/05 22:24:45  rgb
 * Change includes to accomodate RH5.x
 *
 * Revision 1.7  1998/07/29 21:43:17  rgb
 * Convert to 0x-prefixed spis.
 * Support dns lookups for hostnames.
 *
 * Revision 1.6  1998/07/14 18:24:05  rgb
 * Remove unused skbuff header.
 *
 * Revision 1.5  1998/07/09 18:14:11  rgb
 * Added error checking to IP's and keys.
 * Made most error messages more specific rather than spamming usage text.
 * Added more descriptive kernel error return codes and messages.
 * Converted all spi translations to unsigned.
 * Removed all invocations of perror.
 *
 * Revision 1.4  1998/06/30 18:04:32  rgb
 * Fix compiler warning: couldn't find 'struct option' prototype.
 *
 * Revision 1.3  1998/05/27 18:48:20  rgb
 * Adding --help and --version directives.
 *
 * Revision 1.2  1998/05/18 21:14:16  rgb
 * Modifications to be able to ungroup spi's.
 *
 * Revision 1.1.1.1  1998/04/08 05:35:09  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.3  1996/11/20 14:51:32  ji
 * Fixed problems with #include paths.
 * Changed (incorrect) references to ipsp into ipsec.
 *
 * Revision 0.2  1996/11/08 15:46:29  ji
 * First limited release.
 *
 *
 */

#include <sys/types.h>
#include <linux/types.h> /* new */
#include <string.h>
#include <errno.h>

#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
/* #include <linux/ip.h> */

#include <unistd.h>
#include <stdio.h>
#include <netdb.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_netlink.h"
#include "ipsec_ah.h"


u_long bigbuf[1024];
char *program_name;

inline int x2i(char *s)
{
	char ss[3];
	char *endptr;
	long n;
	ss[0] = s[0];
	ss[1] = s[1];
	ss[2] = 0;

	n = strtoul(ss, &endptr, 16);
	if(!(endptr == ss + strlen(ss)))
	{
		fprintf(stderr, "%s: Invalid character in numerical input stream: %s\n",
			program_name, ss);
		exit (1);
	}
	return n;
}

int
resolve_ip(char *arg, struct in_addr *ip)
{
	struct hostent *hostp;
	struct netent *netp;
	
	if (!strcmp(arg, "default")) {/* default is really 0.0.0.0 */
		ip->s_addr = htonl(INADDR_ANY);
		return(1);
	}
	if (!(strcmp(arg, "255.255.255.255"))) {
		/* inet_addr() returns -1 for this, as it should, but that means an error */
		ip->s_addr = htonl(INADDR_BROADCAST);
		return (1);
	}
	if ((netp = getnetbyname(arg)) != (struct netent *)NULL) {
		ip->s_addr = htonl(netp->n_net);
		return(1);
	}
	if ((hostp = gethostbyname(arg)) == (struct hostent *)NULL) {
		errno = h_errno;
		return(-1);
	}
	memcpy((char *) ip, (char *) hostp->h_addr_list[0], hostp->h_length);
	return(0);
}

void
usage(char *s)
{
	fprintf(stderr, "usage: %s dst1 spi1 [ dst2 spi2 [ dst3 spi3 [ dst4 spi4 ] ] ] \n", s);
	exit(1);
}

	

main(argc, argv)
int argc;
char **argv;
{
	int fd;
	int ret;
	int i, dlen, nspis;
	char *endptr;

	struct encap_msghdr *em = (struct encap_msghdr *)bigbuf;

	program_name = argv[0];

	if (argc == 1)
		usage(argv[0]);

        if(strcmp(argv[1], "--help") == 0)
        {
                usage(argv[0]);
        }

        if(strcmp(argv[1], "--version") == 0)
        {
                printf("%s, %s\n", argv[0], spigrp_c_version);
                exit(1);
        }

	if ((argc < 3) || (argc > 9) || ((argc % 2) != 1))
		usage(argv[0]);

	nspis = argc / 2;

	em->em_magic = EM_MAGIC;
	em->em_version = 0;
	if(nspis == 1)
		em->em_type = EMT_UNGRPSPIS;
	else
		em->em_type = EMT_GRPSPIS;
	em->em_msglen = EM_MINLEN + nspis * 12;

	switch (nspis)
	{
	case 4:
		em->em_rel[3].emr_spi = htonl(strtoul(argv[8], &endptr, 0));
		if(!(endptr == argv[8] + strlen(argv[8])))
		{
			fprintf(stderr, "%s: Badly formed spi: %s\n",
				argv[0], argv[8]);
			exit(1);
		}
/*		em->em_rel[3].emr_dst.s_addr = inet_addr(argv[7]); */
#if 0
		if(inet_aton(argv[7], &(em->em_rel[3].emr_dst)) == 0)
		{
			fprintf(stderr, "%s: Badly formed address: %s\n",
				argv[0], argv[7]);
			exit(1);
		}
#endif
		if(ret = resolve_ip(argv[7] /*optarg*/, &(em->em_rel[3].emr_dst)) != 0) {
			if(ret == 1) {
				fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
					program_name, argv[7]/*optarg*/);
				exit (1);
			} else {
				fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
					program_name, argv[7]/*optarg*/);
				exit (1);
			}
		}
	case 3:
		em->em_rel[2].emr_spi = htonl(strtoul(argv[6], &endptr, 0));
		if(!(endptr == argv[6] + strlen(argv[6])))
		{
			fprintf(stderr, "%s: Badly formed spi: %s\n",
				argv[0], argv[6]);
			exit(1);
		}
/*		em->em_rel[2].emr_dst.s_addr = inet_addr(argv[5]); */
#if 0
		if(inet_aton(argv[5], &(em->em_rel[2].emr_dst)) == 0)
		{
			fprintf(stderr, "%s: Badly formed address: %s\n",
				argv[0], argv[5]);
			exit(1);
		}
#endif
		if(ret = resolve_ip(argv[5] /*optarg*/, &(em->em_rel[2].emr_dst)) != 0) {
			if(ret == 1) {
				fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
					program_name, argv[5]/*optarg*/);
				exit (1);
			} else {
				fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
					program_name, argv[5]/*optarg*/);
				exit (1);
			}
		}
	case 2:
		em->em_rel[1].emr_spi = htonl(strtoul(argv[4], &endptr, 0));
		if(!(endptr == argv[4] + strlen(argv[4])))
		{
			fprintf(stderr, "%s: Badly formed spi: %s\n",
				argv[0], argv[4]);
			exit(1);
		}
/*		em->em_rel[1].emr_dst.s_addr = inet_addr(argv[3]); */
#if 0
		if(inet_aton(argv[3], &(em->em_rel[1].emr_dst)) == 0)
		{
			fprintf(stderr, "%s: Badly formed address: %s\n",
				argv[0], argv[3]);
			exit(1);
		}
#endif
		if(ret = resolve_ip(argv[3] /*optarg*/, &(em->em_rel[1].emr_dst)) != 0) {
			if(ret == 1) {
				fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
					program_name, argv[3]/*optarg*/);
				exit (1);
			} else {
				fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
					program_name, argv[3]/*optarg*/);
				exit (1);
			}
		}
	case 1:
		em->em_rel[0].emr_spi = htonl(strtoul(argv[2], &endptr, 0));
		if(!(endptr == argv[2] + strlen(argv[2])))
		{
			fprintf(stderr, "%s: Badly formed spi: %s\n",
				argv[0], argv[2]);
			exit(1);
		}
/*		em->em_rel[0].emr_dst.s_addr = inet_addr(argv[1]); */
#if 0
		if(inet_aton(argv[1], &(em->em_rel[0].emr_dst)) == 0)
		{
			fprintf(stderr, "%s: Badly formed address: %s\n",
				argv[0], argv[1]);
			exit(1);
		}
#endif
		if(ret = resolve_ip(argv[1] /*optarg*/, &(em->em_rel[0].emr_dst)) != 0) {
			if(ret == 1) {
				fprintf(stderr, "%s: Destination address cannot be a network: %s\n",
					program_name, argv[1]/*optarg*/);
				exit (1);
			} else {
				fprintf(stderr, "%s: Invalid formation of destination address: %s\n",
					program_name, argv[1]/*optarg*/);
				exit (1);
			}
		}
		break;
	}
	
	fd = open("/dev/ipsec", 2);
	if (fd < 0)
	{
		printf("%s: Could not open /dev/ipsec -- ", argv[0]);
		switch(errno)
		{
		case ENOENT:
			printf("device does not exist.  See FreeS/WAN installation procedure.\n");
		case EACCES:
			printf("access denied.  ");
			if(getuid() == 0)
			{
				printf("Check permissions.  Should be 600.\n");
			}
			else
			{
				printf("You must be root to open this file.\n");
			}
			break;
		case EUNATCH:
			printf("Netlink not enabled OR KLIPS not loaded.\n");
			break;
		default:
			printf("Unknown file open error %d\n", errno);
		}
		exit(1);
	}
	if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen)
	{
		close(fd);
		printf("%s: Had trouble writing to /dev/ipsec -- ", argv[0]);
		switch(errno)
		{
		case EINVAL:
			printf("Invalid argument, check kernel log messages for specifics.\n");
			break;
		case ENXIO:
			printf("SA not found.  Check kernel log messages for specifics.\n");
			break;
		case EBUSY:
			printf("SA in use.  Check kernel log messages for specifics.\n");
			break;
		default:
			printf("Unknown file write error %d\n", errno);
		}
		close(fd);
		exit(1);
	}
	close(fd);
	exit(0);
}
