/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

/*
 * This file defines various constant values.
 *
 * NOTE:For debugging purposes, constants.c has tables to map numbers back to names.
 * Any changes here should be reflected there.
 */

#define elemsof(array) (sizeof(array) / sizeof(*(array)))	/* number of elements in an array */

typedef int bool;
#define FALSE	0
#define TRUE	1

#define BITS_PER_BYTE	8

/* Routines to check and display values.
 *
 * An enum_desc describes an enumeration.
 * enum_name() returns the name of an enum value, or NULL if invalid.
 * enum_show() is like enum_name, except it formats a numeric representation
 *    for any invalid value (in a static area!)
 *
 * bitnames() formats a display of a set of named bits (in a static area)
 */

typedef const struct enum_desc enum_desc;

extern const char *enum_name(enum_desc *ed, unsigned long val);
extern const char *enum_show(enum_desc *ed, unsigned long val);

extern bool testset(const char *const table[], unsigned long val);
extern const char *bitnamesof(const char *const table[], unsigned long val);


#define FULL_INET_ADDRESS_SIZE    6

/* Group parameters from draft-ietf-ipsec-isakmp-oakley-06.txt section 6 */

#define MODP_GENERATOR "2"

#define MODP768_MODULUS \
    "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1 " \
    "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD " \
    "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245 " \
    "E485B576 625E7EC6 F44C42E9 A63A3620 FFFFFFFF FFFFFFFF"

#define MODP1024_MODULUS \
    "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1 " \
    "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD " \
    "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245 " \
    "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED " \
    "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE65381 " \
    "FFFFFFFF FFFFFFFF"


#define LOCALSECRETSIZE		(256 / BITS_PER_BYTE)

#define DEFAULTNONCESIZE	(128 / BITS_PER_BYTE)

#define MINIMUM_PUBLIC_VALUE_SIZE 32

#define COOKIE_SIZE 8

#define MD5_DIGEST_SIZE		(128 / BITS_PER_BYTE)	/* ought to be supplied by md5.h */
#define SHA1_DIGEST_SIZE	(160 / BITS_PER_BYTE)	/* ought to be supplied by sha1.h */

#define DES_CBC_BLOCK_SIZE	(64 / BITS_PER_BYTE)

/* draft-ietf-ipsec-auth-hmac-sha196-01.txt section 3 */
#define HMAC_SHA1_KEY_LEN    SHA1_DIGEST_SIZE

/* draft-ietf-ipsec-auth-hmac-md5-96-01.txt section 3 */
#define HMAC_MD5_KEY_LEN    MD5_DIGEST_SIZE

#define IKE_UDP_PORT	500

#ifndef PORT
# define PORT IKE_UDP_PORT
#endif

/* Timer events */

extern enum_desc timer_event_desc;

#define EVENT_NULL                              0 /* non-event */
#define EVENT_REINIT_SECRET                     1 /* Refresh cookie secret */
#define EVENT_RETRANSMIT                        2 /* Retransmit packet */
#define EVENT_CLEANUP				3 /* Cleanup responder state */
#define EVENT_SA_EXPIRE				4 /* SA expiration event */

#define EVENT_REINIT_SECRET_DELAY               3600 /* 1 hour */
#define EVENT_RETRANSMIT_DELAY                  30   /* Half a minute */
#define EVENT_CLEANUP_DELAY			120  /* 2 minutes */

#define DEFAULT_SA_EXPIRE                       28800 /* 8 hours */

/* Misc. stuff */

#define SECRET_VALUE_LENGTH  20	/* This is also the size of an SHA-1 digest */

#define ISAKMP_RESPONDER     0
#define ISAKMP_INITIATOR     1

#define MAXIMUM_RETRANSMISSIONS              2

#define UDP_SIZE             65536

#define ZERO_FILE            "/dev/zero"
#define LINUX_RANDOM_FILE    "/dev/urandom"

/* Version numbers */

#define ISAKMP_MAJOR_VERSION   0x1
#define ISAKMP_MINOR_VERSION   0x0

extern enum_desc version_desc;

/* Domain of Interpretation */

extern enum_desc doi_desc;

#define ISAKMP_DOI_ISAKMP          0
#define ISAKMP_DOI_IPSEC           1

/* IPsec DOI things */

#define IPSEC_DOI_SITUATION_LENGTH 4
#define IPSEC_DOI_LDI_LENGTH       4
#define IPSEC_DOI_SPI_SIZE         4

/* debugging settings: a set selections for reporting
 * These would be more naturally situated in log.h,
 * but they are shared with whack.
 */
#ifdef DEBUG
extern const char *const debug_bit_names[];

#define DBG_RAW		0x01	/* raw packet I/O */
#define DBG_CRYPT	0x02	/* encryption/decryption of messages */
#define DBG_PARSING	0x04	/* show decoding of messages */
#define DBG_EMITTING	0x08	/* show encoding of messages */
#define DBG_CONTROL	0x10	/* control flow within Pluto */
#define DBG_KLIPS	0x20	/* messages to KLIPS */

#define DBG_NONE	0	/* no options on */
#define DBG_ALL		0x3f	/* all options on */
#endif

/* State of exchanges
 *
 * The actual value stored in state describes the last message sent,
 * not the message currently being input or output (except during
 * retry).
 *
 * OAKLEY_QUICK_R_0 is an intermediate state (not retained between
 * messages) representing the state when the first quick mode
 * message has been read but not processed.  The structure of the
 * code does not require an analogous OAKLEY_MAIN_R_0.
 *
 * OAKLEY_MAIN_I_4 is an exception because there is no such message.
 * It is best to think of this as a notional empty message following
 * the receipt of OAKLEY_MAIN_R_3 message.
 *
 * Similarly, there is no OAKLEY_QUICK_R_2 message, but the state
 * represents having received the OAKLEY_QUICK_I_2 message.
 *
 * state                 routine that sets it
 * =====                 ====================
 * OAKLEY_MAIN_I_1       main_out_I1
 * OAKLEY_MAIN_R_1       main_inI1_outR1
 * OAKLEY_MAIN_I_2       main_inR1_outI2
 * OAKLEY_MAIN_R_2       main_inI2_outR2
 * OAKLEY_MAIN_I_3       main_inR2_outI3
 * OAKLEY_MAIN_R_3       main_inI3_outR3
 * OAKLEY_MAIN_I_4       main_inR3
 *
 * OAKLEY_QUICK_R_0      comm_handle: encrypted packet but no state
 * OAKLEY_QUICK_I_1      quick_outI1
 * OAKLEY_QUICK_R_1      quick_inI1_outR1
 * OAKLEY_QUICK_I_2      quick_inR1_outI2
 * OAKLEY_QUICK_R_2      quick_inI2
 */

extern enum_desc state_desc;

#define OAKLEY_MAIN_I_1       1
#define OAKLEY_MAIN_R_1       2
#define OAKLEY_MAIN_I_2       3
#define OAKLEY_MAIN_R_2       4
#define OAKLEY_MAIN_I_3       5
#define OAKLEY_MAIN_R_3       6
#define OAKLEY_MAIN_I_4       7

#define OAKLEY_QUICK_R_0      8
#define OAKLEY_QUICK_I_1      9
#define OAKLEY_QUICK_R_1      10
#define OAKLEY_QUICK_I_2      11
#define OAKLEY_QUICK_R_2      12

/* Payload types (draft-ietf-ipsec-isakmp-09.txt section 3.1) */

extern enum_desc payload_desc;

#define ISAKMP_NEXT_NONE       0	/* No other payload following */
#define ISAKMP_NEXT_SA         1	/* Security Association */
#define ISAKMP_NEXT_P          2	/* Proposal */
#define ISAKMP_NEXT_T          3	/* Transform */
#define ISAKMP_NEXT_KE         4	/* Key Exchange */
#define ISAKMP_NEXT_ID         5	/* Identification */
#define ISAKMP_NEXT_CERT       6	/* Certificate */
#define ISAKMP_NEXT_CR         7	/* Certificate Request */
#define ISAKMP_NEXT_HASH       8	/* Hash */
#define ISAKMP_NEXT_SIG        9	/* Signature */
#define ISAKMP_NEXT_NONCE      10	/* Nonce */
#define ISAKMP_NEXT_N          11	/* Notification */
#define ISAKMP_NEXT_D          12       /* Delete */
#define ISAKMP_NEXT_VID        13       /* Vendor ID */

/* Exchange types */

extern enum_desc exchange_desc;

#define ISAKMP_XCHG_NONE       0
#define ISAKMP_XCHG_BASE       1
#define ISAKMP_XCHG_IDPROT     2	/* ID Protection */
#define ISAKMP_XCHG_AO         3	/* Authentication Only */
#define ISAKMP_XCHG_AGGR       4	/* Aggressive */
#define ISAKMP_XCHG_INFO       5	/* Informational */

/* Extra exchange types, defined by Oakley
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */
#define ISAKMP_XCHG_QUICK      32	/* Oakley Quick Mode */
#define ISAKMP_XCHG_NGRP       33	/* Oakley New Group Mode */

/* Flag bits */

extern const char *const flag_bit_names[];

#define ISAKMP_FLAG_ENCRYPTION   0x1
#define ISAKMP_FLAG_COMMIT       0x2

/* Situation definition for IPsec DOI */

extern const char *const sit_bit_names[];

#define SIT_IDENTITY_ONLY        0x01
#define SIT_SECRECY              0x02
#define SIT_INTEGRITY            0x04

/* Protocol IDs (draft-ietf-ipsec-ipsec-doi-08.txt 4.4.1) */

extern enum_desc protocol_desc;

#define PROTO_ISAKMP             1
#define PROTO_IPSEC_AH           2
#define PROTO_IPSEC_ESP          3
#define PROTO_IPCOMP             4

/* warning: trans_show uses enum_show, so same static buffer is used */
#define trans_show(p, t) \
    ((p)==PROTO_IPSEC_AH ? enum_show(&ah_transformid_desc, (t)) \
    : (p)==PROTO_IPSEC_ESP ? enum_show(&esp_transformid_desc, (t)) \
    : "??")

/* IPsec ISAKMP transform values */

extern enum_desc isakmp_transformid_desc;

#define KEY_IKE               1

/* IPsec AH transform values
 * draft-ietf-ipsec-ipsec-doi-07.txt 4.4.3
 */

extern enum_desc ah_transformid_desc;

#define AH_MD5                   2
#define AH_SHA                   3
#define AH_DES                   4

/* IPsec ESP transform values */

extern enum_desc esp_transformid_desc;

#define ESP_NULL                 0
#define ESP_DES_IV64             1
#define ESP_DES                  2
#define ESP_3DES                 3
#define ESP_RC5                  4
#define ESP_IDEA                 5
#define ESP_CAST                 6
#define ESP_BLOWFISH             7
#define ESP_3IDEA                8
#define ESP_DES_IV32             9
#define ESP_ARCFOUR              10

/* IPCOMP transform values */

extern enum_desc ipcomp_desc;

#define IPCOMP_OUI               1
#define IPCOMP_DEFLAT            2
#define IPCOMP_LZS               3
#define IPCOMP_V42BIS            4

/* Identification type values */

extern enum_desc ident_desc;

#define ID_IPV4_ADDR                1
#define ID_FQDN                     2
#define ID_USER_FQDN                3
#define ID_IPV4_ADDR_SUBNET         4
#define ID_IPV6_ADDR                5
#define ID_IPV6_ADDR_SUBNET         6
#define ID_IPV4_ADDR_RANGE          7
#define ID_IPV6_ADDR_RANGE          8
#define ID_DER_ASN1_DN              9
#define ID_DER_ASN1_GN              10
#define ID_KEY_ID                   11

/* Goals for establishing an SA */

extern const char *const sa_goal_bit_names[];

#define GOAL_ENCRYPT                0x01
#define GOAL_AUTHENTICATE           0x02
#define GOAL_TUNNEL                 0x04
#define GOAL_ACCEPT                 0x08

/* Oakley transform attributes
 * draft-ietf-ipsec-isakmp-oakley-07.txt appendix A
 */

extern enum_desc oakley_attr_desc;

#define OAKLEY_ENCRYPTION_ALGORITHM    1
#define OAKLEY_HASH_ALGORITHM          2
#define OAKLEY_AUTHENTICATION_METHOD   3
#define OAKLEY_GROUP_DESCRIPTION       4
#define OAKLEY_GROUP_TYPE              5
#define OAKLEY_GROUP_PRIME             6	/* B/V */
#define OAKLEY_GROUP_GENERATOR_ONE     7	/* B/V */
#define OAKLEY_GROUP_GENERATOR_TWO     8	/* B/V */
#define OAKLEY_GROUP_CURVE_A           9	/* B/V */
#define OAKLEY_GROUP_CURVE_B          10	/* B/V */
#define OAKLEY_LIFE_TYPE              11
#define OAKLEY_LIFE_DURATION          12	/* B/V */
#define OAKLEY_PRF                    13
#define OAKLEY_KEY_LENGTH             14
#define OAKLEY_FIELD_SIZE             15
#define OAKLEY_GROUP_ORDER            16	/* B/V */

/* for each Oakley attribute, which enum_desc describes its values? */
extern enum_desc *oakley_attr_val_descs[];

/* IPsec DOI attributes (draft-ietf-ipsec-ipsec-doi-08.txt 4.5) */

extern enum_desc ipsec_attr_desc;

#define SA_LIFE_TYPE             1
#define SA_LIFE_DURATION         2	/* B/V */
#define GROUP_DESCRIPTION        3
#define ENCAPSULATION_MODE       4
#define AUTH_ALGORITHM           5
#define KEY_LENGTH               6
#define KEY_ROUNDS               7
#define COMPRESS_DICT_SIZE       8
#define COMPRESS_PRIVATE_ALG     9	/* B/V */

/* for each IPsec attribute, which enum_desc describes its values? */
extern enum_desc *ipsec_attr_val_descs[];

/* SA Lifetime Type attribute */

extern enum_desc sa_lifetime_desc;

#define SA_LIFE_TYPE_SECONDS   1
#define SA_LIFE_TYPE_KBYTES    2

#define SA_LIFE_DURATION_DEFAULT    28800 /* Eight hours */
#define SA_LIFE_DURATION_MINIMUM    600   /* 10 minutes */
#define SA_LIFE_DURATION_MAXIMUM    86400 /* One day */

#define SA_LIFE_DURATION_K_MINIMUM  0lu
#define SA_LIFE_DURATION_K_DEFAULT  0xFFFFFFFFlu
#define SA_LIFE_DURATION_K_MAXIMUM  0xFFFFFFFFlu

/* Encapsulation Mode attribute */

extern enum_desc enc_mode_desc;

#define ENCAPSULATION_MODE_UNSPECIFIED 0	/* not legal -- used internally */
#define ENCAPSULATION_MODE_TUNNEL      1
#define ENCAPSULATION_MODE_TRANSPORT   2

/* Auth Algorithm attribute */

extern enum_desc auth_alg_desc;

#define AUTH_ALGORITHM_NONE        0	/* our private designation */
#define AUTH_ALGORITHM_HMAC_MD5    1
#define AUTH_ALGORITHM_HMAC_SHA1   2
#define AUTH_ALGORITHM_DES_MAC     3
#define AUTH_ALGORITHM_KPDK        4

/* Oakley Lifetime Type attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */
extern enum_desc oakley_lifetime_desc;

#define OAKLEY_LIFE_SECONDS   1
#define OAKLEY_LIFE_KILOBYTES 2

#define OAKLEY_ISAKMP_SA_LIFETIME         3600    /* One hour */
#define OAKLEY_ISAKMP_SA_LIFETIME_MINIMUM 1200    /* 20 minutes */
#define OAKLEY_ISAKMP_SA_LIFETIME_MAXIMUM 28800   /* 8 hours */

/* Oakley PRF attribute (none defined)
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */
extern enum_desc oakley_prf_desc;

/* HMAC (see rfc2104.txt) */

#define HMAC_IPAD            0x36
#define HMAC_OPAD            0x5C
#define HMAC_BUFSIZE         64

/* Oakley Encryption Algorithm attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */

extern enum_desc oakley_enc_desc;

#define OAKLEY_DES_CBC          1
#define OAKLEY_IDEA_CBC         2
#define OAKLEY_BLOWFISH_CBC     3
#define OAKLEY_RC5_R16_B64_CBC  4
#define OAKLEY_3DES_CBC         5
#define OAKLEY_CAST_CBC         6

/* Oakley Hash Algorithm attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */

extern enum_desc oakley_hash_desc;

#define OAKLEY_MD5      1
#define OAKLEY_SHA      2
#define OAKLEY_TIGER    3

/* Oakley Authentication Method attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */

extern enum_desc oakley_auth_desc;

#define OAKLEY_PRESHARED_KEY       1
#define OAKLEY_DSS_SIG             2
#define OAKLEY_RSA_SIG             3
#define OAKLEY_RSA_ENC             4
#define OAKLEY_RSA_ENC_REV         5

/* Oakley Group Description attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */
extern enum_desc oakley_group_desc;

#define OAKLEY_GROUP_NONE          0	/* Not legal -- used internally */
#define OAKLEY_GROUP_MODP768       1
#define OAKLEY_GROUP_MODP1024      2
#define OAKLEY_GROUP_GP155         3
#define OAKLEY_GROUP_GP185         4

/* Oakley Group Type attribute
 * draft-ietf-ipsec-isakmp-oakley-06.txt appendix A
 */
extern enum_desc oakley_group_type_desc;

#define OAKLEY_GROUP_TYPE_MODP     1
#define OAKLEY_GROUP_TYPE_ECP      2
#define OAKLEY_GROUP_TYPE_EC2N     3


/* Notify messages -- error types */

extern enum_desc notification_desc;

#define INVALID_PAYLOAD_TYPE     1
#define DOI_NOT_SUPPORTED        2
#define SITUATION_NOT_SUPPORTED  3
#define INVALID_COOKIE           4
#define INVALID_MAJOR_VERSION    5
#define INVALID_MINOR_VERSION    6
#define INVALID_EXCHANGE_TYPE    7
#define INVALID_FLAGS            8
#define INVALID_MESSAGE_ID       9
#define INVALID_PROTOCOL_ID      10
#define INVALID_SPI              11
#define INVALID_TRANSFORM_ID     12
#define ATTRIBUTES_NOT_SUPPORTED 13
#define NO_PROPOSAL_CHOSEN       14
#define BAD_PROPOSAL_SYNTAX      15
#define PAYLOAD_MALFORMED        16
#define INVALID_KEY_INFORMATION  17
#define INVALID_ID_INFORMATION   18
#define INVALID_CERT_ENCODING    19
#define INVALID_CERTIFICATE      20
#define BAD_CERT_REQUEST_SYNTAX  21
#define INVALID_CERT_AUTHORITY   22
#define INVALID_HASH_INFORMATION 23
#define AUTHENTICATION_FAILED    24
#define INVALID_SIGNATURE        25
#define ADDRESS_NOTIFICATION     26
