/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * This file is:
 * Copyright (C) 1998 D. Hugh Redelmeier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This file was written while the author was in Canada.
 *
 * You may contact the author by e-mail: hugh@mimosa.com
 */

/* MODP group constants */

#include <gmp.h>    /* GNU MP library */

extern MP_INT modp768_modulus, modp1024_modulus;
extern MP_INT groupgenerator;

extern void init_crypto(void);


/* unification of cryptographic encoding/decoding algorithms */

#define MAX_OAKLEY_KEY_LEN  (3 * DES_CBC_BLOCK_SIZE)

struct state;	/* forward declaration, dammit */

struct encrypt_desc {
    size_t blocksize;
    size_t keysize;
    void (*crypt)(bool enc, void *buf, size_t buf_len, struct state *st);
};

const struct encrypt_desc oakley_encrypter[OAKLEY_CAST_CBC + 1];


/* unification of cryptographic hashing mechanisms */

union hash_ctx {
	MD5_CTX ctx_md5;
	SHA1_CTX ctx_sha1;
    };

#define MAX_DIGEST_LEN (MD5_DIGEST_SIZE > SHA1_DIGEST_SIZE? MD5_DIGEST_SIZE : SHA1_DIGEST_SIZE)

struct hash_desc {
    size_t hash_digest_len; /* length of digest */
    void (*hash_init)(union hash_ctx *);	/* initialize context */
    void (*hash_update)(union hash_ctx *, const u_char *input, unsigned int len);   /* add input to hash */
    void (*hash_final)(u_char *output, union hash_ctx *);   /* finalize hash */
};

extern const struct hash_desc oakley_hasher[OAKLEY_TIGER+1];


/* HMAC package
 * Note that hmac_ctx can be (and is) copied since there are
 * no persistent pointers into it.
 */

struct hmac_ctx {
    const struct hash_desc *h;	/* underlying hash function */
    size_t hmac_digest_len;	/* copy of h->hash_digest_len */
    union hash_ctx hash_ctx;	/* ctx for hash function */
    u_char buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    };

extern void hmac_init(
    struct hmac_ctx *ctx,
    const struct hash_desc *h,
    const u_char *key,
    size_t key_len);

extern void hmac_reinit(struct hmac_ctx *ctx);	/* saves recreating pads */

extern void hmac_update(
    struct hmac_ctx *ctx,
    const u_char *data,
    size_t data_len);

extern void hmac_final(u_char *output, struct hmac_ctx *ctx);
