/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * This file is:
 * Copyright (C) 1998 D. Hugh Redelmeier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This file was written while the author was in Canada.
 *
 * You may contact the author by e-mail: hugh@mimosa.com
 */

extern bool log_to_stderr;

extern void init_log(void);
extern void log(char *message, ...);
extern void log_errno(char *message, ...);
extern void exit_log(char *message, ...);
extern void exit_log_errno(char *message, ...);

/* Convert struct sockaddr or struct sockaddr_in to a string for presentation.
 * The resulting string is stored in a statically allocated variable,
 * so the next call will overwrite the string.
 * The prototypes use struct sockaddr and struct sockaddr.
 * To ensure that these types are not given prototype scope,
 * we need to declare them.  According to ANSI, the following should work:
 *  struct sockaddr;
 *  struct sockaddr_in;
 * Unfortunately, this isn't good enough for gcc, so we include these
 * headers instead (overkill).
 */
#include <sys/socket.h>
#include <netinet/in.h>

#define SOCKADDR_STRING_SIZE 50	/* should be enough */

extern char *get_address(struct sockaddr sa);
extern char *get_address_in(struct sockaddr_in sin);

extern int get_port(struct sockaddr sa);
extern int get_port_in(struct sockaddr_in sin);

extern const char *show_sa(const struct sockaddr *sa);
extern const char *show_sa_in(const struct sockaddr_in *sin);

#ifdef DEBUG

extern unsigned int debugging;	/* bits selecting what to report */

#define DBG(cond, action)   { if (debugging & (cond)) { action ; } }

extern void DBG_log(char *message, ...);
extern void DBG_dump(const char *label, const void *p, size_t len);
#define DBG_cond_dump(cond, label, p, len) DBG(cond, DBG_dump(label, p, len))

#else

#define DBG(cond, action)	/* do nothing */
#define DBG_cond_dump(cond, label, p, len) /* do nothing */

#endif
